import numpy as np
import pytest

import ross as rs
from ross.materials import steel


def rotor_example1(n_el=48):
    """
    This function instantiate a rotor similar to the example  5.9.1, page 206 (Dynamics of rotating machine, FRISSWELL)
    The following functions test_example1_w_equals0rpm() and test_example1_w_equals4000rpm() are the test functions of
    this example.

    P.S.:Isotropic bearings.

    :param frequency: speed of rotation.
    :param n_el: number of shaft elements.
    :return: Rotor object.
    """

    shaft_elm = []
    for i in range(n_el):
        shaft_elm.append(
            rs.ShaftElement(L=1.5 / n_el, material=steel, n=i, idl=0, odl=0.05)
        )
    disk0 = rs.DiskElement.from_geometry(
        n=(n_el / 1.5) * 0.5, material=steel, width=0.07, i_d=0.05, o_d=0.28
    )
    disk1 = rs.DiskElement.from_geometry(
        n=(n_el / 1.5), material=steel, width=0.07, i_d=0.05, o_d=0.35
    )

    bearing0 = rs.BearingElement(n=0, kxx=1e6, kyy=1e6, cxx=0, cyy=0)
    bearing1 = rs.BearingElement(n=n_el, kxx=1e6, kyy=1e6, cxx=0, cyy=0)

    return rs.Rotor(shaft_elm, [disk0, disk1], [bearing0, bearing1])


def test_example1_w_equals_0rpm():
    rotor = rotor_example1(6)
    modal = rotor.run_modal(0)
    assert modal.wn == pytest.approx(
        np.array([86.66, 86.66, 274.31, 274.31, 716.78, 716.78]), 1e-3
    )


def test_example1_w_equals_4000rpm():
    rotor = rotor_example1(6)
    modal = rotor.run_modal(4000 * np.pi / 30)
    assert modal.wn == pytest.approx(
        np.array([85.39, 87.80, 251.78, 294.71, 600.18, 827.08]), 1e-3
    )


def rotor_example2(n_el=48):
    """
    This function instantiate a overhung rotor similar to the example  5.9.9, page 218 (Dynamics of rotating machine,
    FRISSWELL).
    The following functions test_example2_w_equals0rpm() and test_example2_w_equals4000rpm() are the test functions of
    this example.

    P.S.: Overhung rotor.

    :param n_el: number of shaft elements.
    :return: Rotor object.
    """

    shaft_elm = []
    for i in range(n_el):
        shaft_elm.append(
            rs.ShaftElement(L=1.5 / n_el, material=steel, n=i, idl=0, odl=0.05)
        )
    return rs.Rotor(
        shaft_elm,
        [
            rs.DiskElement.from_geometry(
                n=n_el, material=steel, width=0.07, i_d=0.05, o_d=0.35
            )
        ],
        [
            rs.BearingElement(n=0, kxx=10e6, kyy=10e6, cxx=0, cyy=0),
            rs.BearingElement(n=int((n_el / 1.5)), kxx=10e6, kyy=10e6, cxx=0, cyy=0),
        ],
    )


def test_example2_w_eq_0rpm():
    rotor = rotor_example2(6)
    modal = rotor.run_modal(0)
    assert pytest.approx(modal.wn, 1e-3) == np.array(
        [90.14, 90.14, 630.73, 630.73, 830.43, 830.43]
    )


def test_example2_w_eq_4000rpm():
    rotor = rotor_example2(6)
    modal = rotor.run_modal(4000 * np.pi / 30)
    assert pytest.approx(modal.wn, 1e-3) == np.array(
        [76.19, 103.91, 565.99, 634.23, 647.75, 1174.2]
    )


def rotor_example3(n_el=48):
    """
    This function instantiate a rotor similar to the example  5.9.2, page 208 (Dynamics of rotating machine, FRISSWELL)
    The following functions test_example3_w_equals0rpm() and test_example3_w_equals4000rpm() are the test functions of
    this example.

    P.S.: Anisotropic bearings.

    :param n_el: number of shaft elements.
    :return: Rotor object.
    """
    assert not n_el % 3, "n_el must be a multiple of 3"
    shaft_elm = []
    for i in range(n_el):
        shaft_elm.append(
            rs.ShaftElement(L=1.5 / n_el, material=steel, n=i, idl=0, odl=0.05)
        )
    disk0 = rs.DiskElement.from_geometry(
        n=(n_el / 3), material=steel, width=0.07, i_d=0.05, o_d=0.28
    )
    disk1 = rs.DiskElement.from_geometry(
        n=(n_el / 1.5), material=steel, width=0.07, i_d=0.05, o_d=0.35
    )

    bearing0 = rs.BearingElement(n=0, kxx=1e6, kyy=8e5, cxx=0, cyy=0)
    bearing1 = rs.BearingElement(n=n_el, kxx=1e6, kyy=8e5, cxx=0, cyy=0)

    return rs.Rotor(shaft_elm, [disk0, disk1], [bearing0, bearing1])


def test_example3_w_equals_0rpm():
    rotor = rotor_example3(48)
    modal = rotor.run_modal(0)
    assert pytest.approx(modal.wn, 1e-3) == np.array(
        [82.65, 86.66, 254.52, 274.31, 679.49, 716.79]
    )


def test_example3_w_equals_4000rpm():
    rotor = rotor_example3(48)
    modal = rotor.run_modal(4000 * np.pi / 30)
    assert pytest.approx(modal.wn, 1e-3) == np.array(
        [82.33, 86.86, 239.64, 287.25, 583.49, 806.89]
    )


def rotor_example4(n_el=48):
    """
    This function instantiate a rotor similar to the example  5.9.5, page 212 (Dynamics of rotating machine, FRISSWELL)
    The following functions test_example4_w_equals0rpm() and test_example4_w_equals4000rpm() are the test functions of
    this example.

    P.S.: Isotropic, damped bearings.

    :param n_el: number of shaft elements.
    :return: Rotor object.
    """
    shaft_elm = []
    for i in range(n_el):
        shaft_elm.append(
            rs.ShaftElement(L=1.5 / n_el, material=steel, n=i, idl=0, odl=0.05)
        )
    disk0 = rs.DiskElement.from_geometry(
        n=(n_el / 1.5) * 0.5, material=steel, width=0.07, i_d=0.05, o_d=0.28
    )
    disk1 = rs.DiskElement.from_geometry(
        n=(n_el / 1.5), material=steel, width=0.07, i_d=0.05, o_d=0.35
    )

    bearing0 = rs.BearingElement(n=0, kxx=1e6, kyy=1e6, cxx=3e3, cyy=3e3)
    bearing1 = rs.BearingElement(n=n_el, kxx=1e6, kyy=1e6, cxx=3e3, cyy=3e3)

    return rs.Rotor(shaft_elm, [disk0, disk1], [bearing0, bearing1])


def test_example4_w_equals_0rpm():
    rotor = rotor_example4(6)
    modal = rotor.run_modal(0)
    wn_hertz = modal.wn / (2 * np.pi)
    wd_hertz = modal.wd / (2 * np.pi)
    assert pytest.approx(wn_hertz, 1e-2) == np.array(
        [13.91, 13.91, 48.18, 48.18, 137.06, 137.06]
    )
    assert pytest.approx(wd_hertz, 1e-2) == np.array(
        [13.89, 13.89, 46.54, 46.54, 103.22, 103.22]
    )


def test_example4_w_equals_4000rpm():
    rotor = rotor_example4(6)
    modal = rotor.run_modal((4000 * np.pi) / 30)
    wn_hertz = modal.wn / (2 * np.pi)
    wd_hertz = modal.wd / (2 * np.pi)
    assert pytest.approx(wn_hertz, 1e-1) == np.array(
        [13.70, 14.09, 43.61, 52.18, 122.37, 149.81]
    )
    assert pytest.approx(wd_hertz, 1e-1) == np.array(
        [13.68, 14.07, 41.98, 50.65, 104.25, 105.66]
    )


def rotor_example5(n_el=48):
    """
    This function instantiate a rotor similar to the example  5.9.3, page 209 (Dynamics of rotating machine, FRISSWELL)
    The following function, test_example5_w_equals4000rpm() is the test function of
    this example.

    P.S.: Mixed Modes.

    :param n_el: number of shaft elements.
    :return: Rotor object.
    """

    shaft_elm = []
    for i in range(n_el):
        shaft_elm.append(
            rs.ShaftElement(L=1.5 / n_el, material=steel, n=i, idl=0, odl=0.05)
        )
    disk0 = rs.DiskElement.from_geometry(
        n=(n_el / 3), material=steel, width=0.07, i_d=0.05, o_d=0.28
    )
    disk1 = rs.DiskElement.from_geometry(
        n=(n_el / 1.5), material=steel, width=0.07, i_d=0.05, o_d=0.35
    )

    bearing0 = rs.BearingElement(n=0, kxx=1e6, kyy=2e5, cxx=0, cyy=0)
    bearing1 = rs.BearingElement(n=n_el, kxx=1e6, kyy=2e5, cxx=0, cyy=0)

    return rs.Rotor(shaft_elm, [disk0, disk1], [bearing0, bearing1])


def test_example5_w_equals4000rpm():
    rotor = rotor_example5(48)
    modal = rotor.run_modal((4000 * np.pi) / 30)
    wn_hertz = modal.wn / (2 * np.pi)
    assert pytest.approx(wn_hertz, 1) == (
        np.array([8.545, 13.77, 22.35, 44.06, 78.76, 120.4])
    ) / (2 * np.pi)


def rotor_example6(n_el=48):
    """
    This function instantiate a rotor similar to the example  5.9.4, page 210 (Dynamics of rotating machine, FRISSWELL)
    The following functions test_example6_w_equals0rpm() and test_example6_w_equals4000rpm() are the test functions of
    this example.

    P.S.:Cross-coupled bearings.

    :param n_el: number of shaft elements.
    :return: Rotor object.
    """

    shaft_elm = []
    for i in range(n_el):
        shaft_elm.append(
            rs.ShaftElement(L=1.5 / n_el, material=steel, n=i, idl=0, odl=0.05)
        )
    disk0 = rs.DiskElement.from_geometry(
        n=(n_el / 1.5) * 0.5, material=steel, width=0.07, i_d=0.05, o_d=0.28
    )
    disk1 = rs.DiskElement.from_geometry(
        n=(n_el / 1.5), material=steel, width=0.07, i_d=0.05, o_d=0.35
    )

    bearing0 = rs.BearingElement(n=0, kxx=1e6, kyy=1e6, cxx=0, kxy=5e5, cyy=0)
    bearing1 = rs.BearingElement(n=n_el, kxx=1e6, kyy=1e6, cxx=0, kxy=5e5, cyy=0)

    return rs.Rotor(shaft_elm, [disk0, disk1], [bearing0, bearing1])


def test_example6_w_equals4000rpm():
    rotor = rotor_example6(48)
    modal = rotor.run_modal((4000 * np.pi) / 30)
    wn_hertz = modal.wn / (2 * np.pi)
    assert pytest.approx(wn_hertz, 1) == (
        np.array([11.66, 14.80, 33.97, 49.19, 97.97, 126.61])
    )



"""Shaft element module for STOCHASTIC ROSS.

This module creates an instance of random shaft element for stochastic
analysis.
"""

from ross.shaft_element import ShaftElement
from ross.stochastic.st_materials import ST_Material
from ross.stochastic.st_results_elements import plot_histogram
from ross.units import Q_, check_units

__all__ = ["ST_ShaftElement", "st_shaft_example"]


class ST_ShaftElement:
    """Random shaft element.

    Creates an object containing a generator with random instances of
    ShaftElement.

    Parameters
    ----------
    L : float, pint.Quantity, list
        Element length.
        Input a list to make it random.
    idl : float, pint.Quantity, list
        Inner diameter of the element at the left position.
        Input a list to make it random.
    odl : float, pint.Quantity, list
        Outer diameter of the element at the left position.
        Input a list to make it random.
    idr : float, pint.Quantity, list, optional
        Inner diameter of the element at the right position
        Default is equal to idl value (cylindrical element)
        Input a list to make it random.
    odr : float, pint.Quantity, list, optional
        Outer diameter of the element at the right position.
        Default is equal to odl value (cylindrical element)
        Input a list to make it random.
    material : ross.material, list of ross.material
        Shaft material.
        Input a list to make it random.
    n : int, optional
        Element number (coincident with it's first node).
        If not given, it will be set when the rotor is assembled
        according to the element's position in the list supplied to
    axial_force : float, list, optional
        Axial force.
        Input a list to make it random.
        Default is 0.
    torque : float, list, optional
        Torque
        Input a list to make it random.
        Default is 0.
    shear_effects : bool, optional
        Determine if shear effects are taken into account.
        Default is True.
    rotary_inertia : bool, optional
        Determine if rotary_inertia effects are taken into account.
        Default is True.
    gyroscopic : bool, optional
        Determine if gyroscopic effects are taken into account.
        Default is True.
    shear_method_calc : str, optional
        Determines which shear calculation method the user will adopt.
        Default is 'cowper'
    is_random : list
        List of the object attributes to become random.
        Possibilities:
            ["L", "idl", "odl", "idr", "odr", "material", "axial_force", "torque"]

    Example
    -------
    >>> import numpy as np
    >>> import ross.stochastic as srs
    >>> size = 5
    >>> E = np.random.uniform(208e9, 211e9, size)
    >>> st_steel = srs.ST_Material(name="Steel", rho=7810, E=E, G_s=81.2e9)
    >>> elms = srs.ST_ShaftElement(L=1,
    ...                            idl=0,
    ...                            odl=np.random.uniform(0.1, 0.2, size),
    ...                            material=st_steel,
    ...                            is_random=["odl", "material"],
    ...                            )
    >>> len(list(iter(elms)))
    5
    """

    @check_units
    def __init__(
        self,
        L,
        idl,
        odl,
        idr=None,
        odr=None,
        material=None,
        n=None,
        axial_force=0,
        torque=0,
        shear_effects=True,
        rotary_inertia=True,
        gyroscopic=True,
        shear_method_calc="cowper",
        is_random=None,
    ):
        if idr is None:
            idr = idl
            if "idl" in is_random and "idr" not in is_random:
                is_random.append("idr")
        if odr is None:
            odr = odl
            if "odl" in is_random and "odr" not in is_random:
                is_random.append("odr")
        if isinstance(material, ST_Material):
            material = list(iter(material))

        attribute_dict = dict(
            L=L,
            idl=idl,
            odl=odl,
            idr=idr,
            odr=odr,
            material=material,
            n=n,
            axial_force=axial_force,
            torque=torque,
            shear_effects=shear_effects,
            rotary_inertia=rotary_inertia,
            gyroscopic=gyroscopic,
            shear_method_calc=shear_method_calc,
            tag=None,
        )
        self.is_random = is_random
        self.attribute_dict = attribute_dict

    def __iter__(self):
        """Return an iterator for the container.

        Returns
        -------
        An iterator over random shaft elements.

        Examples
        --------
        >>> import ross.stochastic as srs
        >>> elm = srs.st_shaft_example()
        >>> len(list(iter(elm)))
        2
        """
        return iter(self.random_var(self.is_random, self.attribute_dict))

    def __getitem__(self, key):
        """Return the value for a given key from attribute_dict.

        Parameters
        ----------
        key : str
            A class parameter as string.

        Raises
        ------
        KeyError
            Raises an error if the parameter doesn't belong to the class.

        Returns
        -------
        Return the value for the given key.

        Example
        -------
        >>> import numpy as np
        >>> import ross.stochastic as srs
        >>> size = 5
        >>> E = np.random.uniform(208e9, 211e9, size)
        >>> st_steel = srs.ST_Material(name="Steel", rho=7810, E=E, G_s=81.2e9)
        >>> elms = srs.ST_ShaftElement(L=1,
        ...                            idl=0,
        ...                            odl=np.random.uniform(0.1, 0.2, size),
        ...                            material=st_steel,
        ...                            is_random=["odl", "material"],
        ...                            )
        >>> elms["L"]
        1
        """
        if key not in self.attribute_dict.keys():
            raise KeyError("Object does not have parameter: {}.".format(key))

        return self.attribute_dict[key]

    def __setitem__(self, key, value):
        """Set new parameter values for the object.

        Function to change a parameter value.
        It's not allowed to add new parameters to the object.

        Parameters
        ----------
        key : str
            A class parameter as string.
        value : The corresponding value for the attrbiute_dict's key.
            ***check the correct type for each key in ST_ShaftElement
            docstring.

        Raises
        ------
        KeyError
            Raises an error if the parameter doesn't belong to the class.

        Example
        -------
        >>> import numpy as np
        >>> import ross.stochastic as srs
        >>> size = 5
        >>> E = np.random.uniform(208e9, 211e9, size)
        >>> st_steel = srs.ST_Material(name="Steel", rho=7810, E=E, G_s=81.2e9)
        >>> elms = srs.ST_ShaftElement(L=1,
        ...                            idl=0,
        ...                            odl=np.random.uniform(0.1, 0.2, size),
        ...                            material=st_steel,
        ...                            is_random=["odl", "material"],
        ...                            )
        >>> elms["odl"] = np.linspace(0.1, 0.2, 5)
        >>> elms["odl"]
        array([0.1  , 0.125, 0.15 , 0.175, 0.2  ])
        """
        if key not in self.attribute_dict.keys():
            raise KeyError("Object does not have parameter: {}.".format(key))
        self.attribute_dict[key] = value

    def random_var(self, is_random, *args):
        """Generate a list of objects as random attributes.

        This function creates a list of objects with random values for selected
        attributes from ross.ShaftElement.

        Parameters
        ----------
        is_random : list
            List of the object attributes to become stochastic.
        *args : dict
            Dictionary instanciating the ross.ShaftElement class.
            The attributes that are supposed to be stochastic should be
            set as lists of random variables.

        Returns
        -------
        f_list : generator
            Generator of random objects.
        """
        args_dict = args[0]
        new_args = []
        for i in range(len(args_dict[is_random[0]])):
            arg = []
            for key, value in args_dict.items():
                if key in is_random:
                    arg.append(value[i])
                else:
                    arg.append(value)
            new_args.append(arg)
        f_list = (ShaftElement(*arg) for arg in new_args)

        return f_list

    def plot_random_var(self, var_list=None, histogram_kwargs=None, plot_kwargs=None):
        """Plot histogram and the PDF.

        This function creates a histogram to display the random variable
        distribution.

        Parameters
        ----------
        var_list : list, optional
            List of random variables, in string format, to plot.
            Default is plotting all the random variables.
        histogram_kwargs : dict, optional
            Additional key word arguments can be passed to change
            the plotly.go.histogram (e.g. histnorm="probability density", nbinsx=20...).
            *See Plotly API to more information.
        plot_kwargs : dict, optional
            Additional key word arguments can be passed to change the plotly go.figure
            (e.g. line=dict(width=4.0, color="royalblue"), opacity=1.0, ...).
            *See Plotly API to more information.

        Returns
        -------
        fig : Plotly graph_objects.Figure()
            A figure with the histogram plots.

        Examples
        --------
        >>> import ross.stochastic as srs
        >>> elm = srs.st_shaft_example()
        >>> fig = elm.plot_random_var(["odl"])
        >>> # fig.show()
        """
        label = dict(
            L="Length",
            idl="Left inner diameter",
            odl="Left outer diameter",
            idr="Right inner diameter",
            odr="Right outer diameter",
        )
        is_random = self.is_random
        if "material" in is_random:
            is_random.remove("material")

        if var_list is None:
            var_list = is_random
        elif not all(var in is_random for var in var_list):
            raise ValueError(
                "Random variable not in var_list. Select variables from {}".format(
                    is_random
                )
            )

        return plot_histogram(
            self.attribute_dict, label, var_list, histogram_kwargs={}, plot_kwargs={}
        )


def st_shaft_example():
    """Return an instance of a simple random shaft element.

    The purpose is to make available a simple model so that doctest can be
    written using it.

    Returns
    -------
    elm : ross.stochastic.ST_ShaftElement
        An instance of a random shaft element object.

    Examples
    --------
    >>> import ross.stochastic as srs
    >>> elm = srs.st_shaft_example()
    >>> len(list(iter(elm)))
    2
    """
    from ross.materials import steel

    elm = ST_ShaftElement(
        L=[1.0, 1.1],
        idl=0.0,
        odl=[0.1, 0.2],
        material=steel,
        is_random=["L", "odl"],
    )
    return elm



"""Plotting module for elements.

This modules provides functions to plot the elements statistic data.
"""

from copy import copy

import numpy as np
from plotly import graph_objects as go
from plotly import io as pio
from plotly.subplots import make_subplots
from scipy.stats import gaussian_kde

from ross.plotly_theme import tableau_colors


def plot_histogram(
    attribute_dict, label={}, var_list=[], histogram_kwargs=None, plot_kwargs=None
):
    """Plot histogram and the PDF.

    This function creates a histogram to display the random variable distribution.

    Parameters
    ----------
    attribute_dict : dict
        Dictionary with element parameters.
    label : dict
        Dictionary with labels for each element parameter. Labels are displayed
        on plotly figure.
    var_list : list, optional
        List of random variables, in string format, to plot.
    histogram_kwargs : dict, optional
        Additional key word arguments can be passed to change
        the plotly.go.histogram (e.g. histnorm="probability density", nbinsx=20...).
        *See Plotly API to more information.
    plot_kwargs : dict, optional
        Additional key word arguments can be passed to change the plotly go.figure
        (e.g. line=dict(width=4.0, color="royalblue"), opacity=1.0...).
        *See Plotly API to more information.

    Returns
    -------
    subplots : Plotly graph_objects.make_subplots()
        A figure with the histogram plots.
    """
    histogram_kwargs = {} if histogram_kwargs is None else copy(histogram_kwargs)
    plot_kwargs = {} if plot_kwargs is None else copy(plot_kwargs)

    hist_default_values = dict(
        histnorm="probability density",
        cumulative_enabled=False,
        nbinsx=20,
        marker_color=tableau_colors["red"],
        opacity=1.0,
    )
    for k, v in hist_default_values.items():
        histogram_kwargs.setdefault(k, v)

    plot_default_values = dict(
        line=dict(width=4.0, color=tableau_colors["blue"]), opacity=1.0
    )
    for k, v in plot_default_values.items():
        plot_kwargs.setdefault(k, v)

    rows = 1 if len(var_list) < 2 else 2
    cols = len(var_list) // 2 + len(var_list) % 2
    fig = make_subplots(rows=rows, cols=cols)

    for i, var in enumerate(var_list):
        row = i % 2 + 1
        col = i // 2 + 1
        if histogram_kwargs["histnorm"] == "probability density":
            if histogram_kwargs["cumulative_enabled"] is True:
                y_label = "CDF"
            else:
                y_label = "PDF"
        else:
            y_label = "Frequency"

        fig.add_trace(
            go.Histogram(
                x=attribute_dict[var],
                name="Histogram",
                legendgroup="Histogram",
                showlegend=True if i == 0 else False,
                **histogram_kwargs,
            ),
            row=row,
            col=col,
        )
        if y_label == "PDF":
            x = np.linspace(
                min(attribute_dict[var]),
                max(attribute_dict[var]),
                len(attribute_dict[var]),
            )
            kernel = gaussian_kde(attribute_dict[var])
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=kernel(x),
                    mode="lines",
                    name="PDF Estimation",
                    legendgroup="PDF Estimation",
                    showlegend=True if i == 0 else False,
                    **plot_kwargs,
                ),
                row=row,
                col=col,
            )
        fig.update_xaxes(
            title=dict(text="<b>{}</b>".format(label[var])),
            exponentformat="E",
            row=row,
            col=col,
        )
        fig.update_yaxes(
            title=dict(text="<b>{}</b>".format(y_label), standoff=0),
            exponentformat="E",
            row=row,
            col=col,
        )
    fig.update_layout(bargroupgap=0.1, plot_bgcolor="white")

    return fig



"""This module deals with lubricants dictionary in the ROSS library."""

from ross.units import Q_

lubricant_dict = {
    "ISOVG32": {
        "viscosity1": Q_(4.05640e-06, "reyn").to_base_units().m,
        "temp1": Q_(40.00000, "degC").to_base_units().m,
        "viscosity2": Q_(6.76911e-07, "reyn").to_base_units().m,
        "temp2": Q_(100.00000, "degC").to_base_units().m,
        "lube_density": Q_(873.99629, "kg/m³").to_base_units().m,
        "lube_cp": Q_(1948.7995685758851, "J/(kg*degK)").to_base_units().m,
        "lube_conduct": Q_(0.13126, "W/(m*degC)").to_base_units().m,
    },
    "ISOVG46": {
        "viscosity1": Q_(5.757040938820288e-06, "reyn").to_base_units().m,
        "temp1": Q_(40, "degC").to_base_units().m,
        "viscosity2": Q_(8.810775697672788e-07, "reyn").to_base_units().m,
        "temp2": Q_(100, "degC").to_base_units().m,
        "lube_density": Q_(862.9, "kg/m³").to_base_units().m,
        "lube_cp": Q_(1950, "J/(kg*degK)").to_base_units().m,
        "lube_conduct": Q_(0.15, "W/(m*degC)").to_base_units().m,
    },
    "TEST": {
        "viscosity1": Q_(0.04, "Pa*s").to_base_units().m,
        "temp1": Q_(40, "degC").to_base_units().m,
        "viscosity2": Q_(0.01, "Pa*s").to_base_units().m,
        "temp2": Q_(100, "degC").to_base_units().m,
        "lube_density": Q_(863.61302696, "kg/m³").to_base_units().m,
        "lube_cp": Q_(1951.88616, "J/(kg*degK)").to_base_units().m,
        "lube_conduct": Q_(0.15, "W/(m*degC)").to_base_units().m,
    },
    "ISOVG68": {
        "viscosity1": Q_(0.0570722, "Pa*s").to_base_units().m,
        "temp1": Q_(40, "degC").to_base_units().m,
        "viscosity2": Q_(0.00766498, "Pa*s").to_base_units().m,
        "temp2": Q_(100, "degC").to_base_units().m,
        "lube_density": Q_(867, "kg/m³").to_base_units().m,
        "lube_cp": Q_(1951, "J/(kg*degK)").to_base_units().m,
        "lube_conduct": Q_(0.15, "W/(m*degC)").to_base_units().m,
    },
}



"""Time integration module.

This module defines the time integration routines to evaluate time responses from
the rotors. These simulate the transient evolution of the dynamic behavior of the
rotors defined by the user, and return the time signals for rotor displacement.

"""

import numpy as np


class Integrator:
    """A series of Runge-Kutta time integration algorithms.

    Calculates the time response for the rotors input to the routine.

    Parameters
    ----------
    x0 : float
        Initial time
    y0 : float
        Initial condition for the integration
    x : float
        Iteration time
    h : float
        Step height
    func : object
        Function to be integrated in time

    Returns
    -------
    The rotor transient response.

    References
    ----------
    .. [1] BUTCHER, John Charles; GOODWIN, Nicolette. Numerical methods for ordinary differential equations. New York: Wiley, 2008. ..
    .. [2] CASH, Jeff R.; KARP, Alan H. A variable order Runge-Kutta method for initial value problems with rapidly varying right-hand sides.
           ACM Transactions on Mathematical Software (TOMS), v. 16, n. 3, p. 201-222, 1990. ..

    """

    def __init__(self, x0, y0, x, h, func, print_progress=False):
        self.x0 = x0
        self.y0 = y0
        self.x = x
        self.h = h
        self.func = func
        self.print_progress = print_progress

    def rk4(self):
        # Runge-Kutta 4th order (RK4)

        # Count number of iterations using step size or
        # step height h
        n = int((self.x - self.x0) / self.h)

        # Iterate for number of iterations
        y = self.y0
        result = np.zeros((24, n + 1))
        result[:, 0] = self.y0

        # 4th-order Runge-Kutta

        for i in range(1, n + 1):
            if i % 10000 == 0 and self.print_progress:
                print(f"Iteration: {i} \n Time: {self.x0}")

            "Apply Runge Kutta Formulas to find next value of y"
            k1 = self.h * self.func(self.x0, y, i)
            k2 = self.h * self.func(self.x0 + 0.5 * self.h, y + 0.5 * k1, i)
            k3 = self.h * self.func(self.x0 + 0.5 * self.h, y + 0.5 * k2, i)
            k4 = self.h * self.func(self.x0 + self.h, y + k3, i)

            # Update next value of y
            y = y + (1.0 / 6.0) * (k1 + 2 * k2 + 2 * k3 + k4)
            result[:, i] = np.copy(y)

            # Update next value of x
            self.x0 = self.x0 + self.h

        return result

    def rk45(self):
        # Runge-Kutta Cash-Karp (CK45)

        # Count number of iterations using step size or
        # step height h
        n = int((self.x - self.x0) / self.h)

        # Iterate for number of iterations
        y = self.y0
        result = np.zeros((24, n + 1))
        result[:, 0] = self.y0

        for i in range(1, n + 1):
            if i % 10000 == 0 and self.print_progress:
                print(f"Iteration: {i} \n Time: {self.x0}")

            "Apply Runge Kutta Formulas to find next value of y"
            k1 = 1 * self.func(self.x0, y, i)
            yp2 = y + k1 * (self.h / 5)
            k2 = 1 * self.func(self.x0 + (self.h / 5), yp2, i)
            yp3 = y + k1 * (3 * self.h / 40) + k2 * (9 * self.h / 40)
            k3 = 1 * self.func(self.x0 + (3 * self.h / 10), yp3, i)
            yp4 = (
                y
                + k1 * (3 * self.h / 10)
                - k2 * (9 * self.h / 10)
                + k3 * (6 * self.h / 5)
            )
            k4 = 1 * self.func(self.x0 + (3 * self.h / 5), yp4, i)
            yp5 = (
                y
                - k1 * (11 * self.h / 54)
                + k2 * (5 * self.h / 2)
                - k3 * (70 * self.h / 27)
                + k4 * (35 * self.h / 27)
            )
            k5 = 1 * self.func(self.x0 + self.h, yp5, i)
            yp6 = (
                y
                + k1 * (1631 * self.h / 55296)
                + k2 * (175 * self.h / 512)
                + k3 * (575 * self.h / 13824)
                + k4 * (44275 * self.h / 110592)
                + k5 * (253 * self.h / 4096)
            )
            k6 = 1 * self.func(self.x0 + (7 * self.h / 8), yp6, i)

            # Update next value of y
            y = y + self.h * (
                37 * k1 / 378 + 250 * k3 / 621 + 125 * k4 / 594 + 512 * k6 / 1771
            )
            result[:, i] = np.copy(y)

            # Update next value of x
            self.x0 = self.x0 + self.h

        return result



"""Misalignment module.

This module defines misalignments of various types on the shaft coupling. There are 
a number of options, for the formulation of 6 DoFs (degrees of freedom).
"""

import time

import numpy as np
from scipy import linalg as la

import ross
from ross.units import Q_, check_units

from .abs_fault import Fault
from .integrate_solver import Integrator

__all__ = ["MisalignmentFlex", "MisalignmentRigid"]


class MisalignmentFlex(Fault):
    """A flexible coupling with misalignment of some kind.

    Calculates the dynamic reaction force of hexangular flexible coupling
    induced by 6DOF's rotor parallel and angular misalignment.

    Parameters
    ----------
    dt : float
        Time step.
    tI : float
        Initial time.
    tF : float
        Final time.
    kd : float
        Radial stiffness of flexible coupling.
    ks : float
        Bending stiffness of flexible coupling.
    eCOUPx : float
        Parallel misalignment offset between driving rotor and driven rotor along X direction.
    eCOUPy : float
        Parallel misalignment offset between driving rotor and driven rotor along Y direction.
    misalignment_angle : float
        Angular misalignment angle.
    TD : float
        Driving torque.
    TL : float
        Driven torque.
    n1 : float
        Node where the misalignment is ocurring.
    speed : float, pint.Quantity
        Operational speed of the machine. Default unit is rad/s.
    unbalance_magnitude : array
        Array with the unbalance magnitude. The unit is kg.m.
    unbalance_phase : array
        Array with the unbalance phase. The unit is rad.
    mis_type: string
        String containing the misalignment type choosed. The avaible types are: parallel, by default; angular; combined.
    print_progress : bool
        Set it True, to print the time iterations and the total time spent.
        False by default.

    Returns
    -------
    A force to be applied on the shaft.

    References
    ----------
    .. [1] 'Xia, Y., Pang, J., Yang, L., Zhao, Q., & Yang, X. (2019). Study on vibration response
    and orbits of misaligned rigid rotors connected by hexangular flexible coupling. Applied
    Acoustics, 155, 286-296 ..

    Examples
    --------
    >>> from ross.probe import Probe
    >>> from ross.faults.misalignment import misalignment_flex_parallel_example
    >>> probe1 = Probe(14, 0)
    >>> probe2 = Probe(22, 0)
    >>> response = misalignment_flex_parallel_example()
    >>> results = response.run_time_response()
    >>> fig = response.plot_dfft(probe=[probe1, probe2], range_freq=[0, 100], yaxis_type="log")
    >>> # fig.show()
    """

    @check_units
    def __init__(
        self,
        dt,
        tI,
        tF,
        kd,
        ks,
        eCOUPx,
        eCOUPy,
        misalignment_angle,
        TD,
        TL,
        n1,
        speed,
        unbalance_magnitude,
        unbalance_phase,
        mis_type,
        print_progress=False,
    ):
        self.dt = dt
        self.tI = tI
        self.tF = tF
        self.kd = kd
        self.ks = ks
        self.eCOUPx = eCOUPx
        self.eCOUPy = eCOUPy
        self.misalignment_angle = misalignment_angle
        self.TD = TD
        self.TL = TL
        self.n1 = n1
        self.n2 = n1 + 1
        self.speed = speed
        self.unbalance_magnitude = unbalance_magnitude
        self.unbalance_phase = unbalance_phase

        self.speedI = speed
        self.speedF = speed

        self.mis_type = mis_type
        self.print_progress = print_progress

        if self.mis_type is None or self.mis_type == "parallel":
            self._force = self._parallel
        elif self.mis_type == "angular":
            self._force = self._angular
        elif self.mis_type == "combined":
            self._force = self._combined
        else:
            raise Exception("Check the misalignment type!")

        if len(self.unbalance_magnitude) != len(self.unbalance_phase):
            raise Exception(
                "The unbalance magnitude vector and phase must have the same size!"
            )

    def run(self, rotor):
        """Calculates the shaft angular position and the misalignment amount at X / Y directions.

        Parameters
        ----------
        rotor : ross.Rotor Object
             6 DoF rotor model.

        """
        self.rotor = rotor
        self.n_disk = len(self.rotor.disk_elements)
        if self.n_disk != len(self.unbalance_magnitude):
            raise Exception("The number of discs and unbalances must agree!")

        self.radius = rotor.elements[self.n1].odl / 2
        self.ndof = rotor.ndof
        self.ndofd = np.zeros(len(self.rotor.disk_elements))

        for ii in range(self.n_disk):
            self.ndofd[ii] = (self.rotor.disk_elements[ii].n) * 6

        self.Cte = (
            self.ks * self.radius * np.sqrt(2 - 2 * np.cos(self.misalignment_angle))
        )

        # parameters for the time integration
        self.lambdat = 0.00001
        Faxial = 0
        TorqueI = 0
        TorqueF = 0

        # pre-processing of auxilary variuables for the time integration
        self.sA = (
            self.speedI * np.exp(-self.lambdat * self.tF)
            - self.speedF * np.exp(-self.lambdat * self.tI)
        ) / (np.exp(-self.lambdat * self.tF) - np.exp(-self.lambdat * self.tI))
        self.sB = (self.speedF - self.speedI) / (
            np.exp(-self.lambdat * self.tF) - np.exp(-self.lambdat * self.tI)
        )

        # This code below here is used for acceleration and torque application to the rotor. As of
        # september/2020 it is unused, but might be implemented in future releases. These would be
        # run-up and run-down operations and variations of operating conditions.
        #
        # sAT = (
        #     TorqueI * np.exp(-lambdat * self.tF) - TorqueF * np.exp(-lambdat * self.tI)
        # ) / (np.exp(-lambdat * self.tF) - np.exp(-lambdat * self.tI))
        # sBT = (TorqueF - TorqueI) / (
        #     np.exp(-lambdat * self.tF) - np.exp(-lambdat * self.tI)
        # )
        #
        # SpeedV = sA + sB * np.exp(-lambdat * self.t)
        # TorqueV = sAT + sBT * np.exp(-lambdat * self.t)
        # AccelV = -lambdat * sB * np.exp(-lambdat * self.t)

        # Determining the modal matrix
        self.K = self.rotor.K(self.speed)
        self.C = self.rotor.C(self.speed)
        self.G = self.rotor.G()
        self.M = self.rotor.M(self.speed)
        self.Ksdt = self.rotor.Ksdt()

        _, ModMat = la.eigh(self.K, self.M)
        ModMat = ModMat[:, :12]
        self.ModMat = ModMat

        # Modal transformations
        self.Mmodal = ((ModMat.T).dot(self.M)).dot(ModMat)
        self.Cmodal = ((ModMat.T).dot(self.C)).dot(ModMat)
        self.Gmodal = ((ModMat.T).dot(self.G)).dot(ModMat)
        self.Kmodal = ((ModMat.T).dot(self.K)).dot(ModMat)
        self.Ksdtmodal = ((ModMat.T).dot(self.Ksdt)).dot(ModMat)

        # Omega = self.speedI * np.pi / 30

        y0 = np.zeros(24)
        t_eval = np.arange(self.tI, self.tF + self.dt, self.dt)
        T = t_eval

        self.angular_position = (
            self.sA * T
            - (self.sB / self.lambdat) * np.exp(-self.lambdat * T)
            + (self.sB / self.lambdat)
        )

        self.Omega = self.sA + self.sB * np.exp(-self.lambdat * T)
        self.AccelV = -self.lambdat * self.sB * np.exp(-self.lambdat * T)

        self.tetaUNB = np.zeros((len(self.unbalance_phase), len(self.angular_position)))
        unbx = np.zeros(len(self.angular_position))
        unby = np.zeros(len(self.angular_position))

        FFunb = np.zeros((self.ndof, len(t_eval)))

        for ii in range(self.n_disk):
            self.tetaUNB[ii, :] = (
                self.angular_position + self.unbalance_phase[ii] + np.pi / 2
            )

            unbx = self.unbalance_magnitude[ii] * (self.AccelV) * (
                np.cos(self.tetaUNB[ii, :])
            ) - self.unbalance_magnitude[ii] * ((self.Omega**2)) * (
                np.sin(self.tetaUNB[ii, :])
            )

            unby = -self.unbalance_magnitude[ii] * (self.AccelV) * (
                np.sin(self.tetaUNB[ii, :])
            ) - self.unbalance_magnitude[ii] * (self.Omega**2) * (
                np.cos(self.tetaUNB[ii, :])
            )

            FFunb[int(self.ndofd[ii]), :] += unbx
            FFunb[int(self.ndofd[ii] + 1), :] += unby

        self.Funbmodal = (self.ModMat.T).dot(FFunb)

        self.inv_Mmodal = np.linalg.pinv(self.Mmodal)
        t1 = time.time()
        self.forces = self._force(self.angular_position)

        self.ft_modal = (self.ModMat.T).dot(self.forces).T

        x = Integrator(
            self.tI,
            y0,
            self.tF,
            self.dt,
            self._equation_of_movement,
            self.print_progress,
        )
        x = x.rk45()
        t2 = time.time()
        if self.print_progress:
            print(f"Time spent: {t2-t1} s")

        self.displacement = x[:12, :]
        self.velocity = x[12:, :]
        self.time_vector = t_eval
        self.response = self.ModMat.dot(self.displacement)

    def _equation_of_movement(self, T, Y, i):
        """Calculates the displacement and velocity using state-space representation in the modal domain.

        Parameters
        ----------
        T : float
            Iteration time.
        Y : array
            Array of displacement and velocity, in the modal domain.
        i : int
            Iteration step.

        Returns
        -------
        new_Y :  array
            Array of the new displacement and velocity, in the modal domain.
        """

        positions = Y[:12]
        velocity = Y[12:]  # velocity ign space state

        ftmodal = self.ft_modal[i]

        # proper equation of movement to be integrated in time
        new_V_dot = (
            ftmodal
            + self.Funbmodal[:, i]
            - ((self.Cmodal + self.Gmodal * self.Omega[i])).dot(velocity)
            - ((self.Kmodal + self.Ksdtmodal * self.AccelV[i]).dot(positions))
        ).dot(self.inv_Mmodal)

        new_X_dot = velocity

        new_Y = np.zeros(24)
        new_Y[:12] = new_X_dot
        new_Y[12:] = new_V_dot

        return new_Y

    def _parallel(self, angular_position):
        """Reaction forces of parallel misalignment.

        angular_position : float
                        Angular position of the shaft.

        Returns
        -------
        F_mis_p : array
               Excitation force caused by the parallel misalignment on the entire system.
        """

        F_mis_p = np.zeros((self.ndof, len(angular_position)))

        fib = np.arctan(self.eCOUPy / self.eCOUPx)

        self.mi_y = (
            (
                np.sqrt(
                    self.radius**2
                    + self.eCOUPx**2
                    + self.eCOUPy**2
                    + 2
                    * self.radius
                    * np.sqrt(self.eCOUPx**2 + self.eCOUPy**2)
                    * np.sin(fib + angular_position)
                )
                - self.radius
            )
            * np.cos(angular_position)
            + (
                np.sqrt(
                    self.radius**2
                    + self.eCOUPx**2
                    + self.eCOUPy**2
                    + 2
                    * self.radius
                    * np.sqrt(self.eCOUPx**2 + self.eCOUPy**2)
                    * np.cos(np.pi / 6 + fib + angular_position)
                )
                - self.radius
            )
            * np.cos(2 * np.pi / 3 + angular_position)
            + (
                self.radius
                - np.sqrt(
                    self.radius**2
                    + self.eCOUPx**2
                    + self.eCOUPy**2
                    - 2
                    * self.radius
                    * np.sqrt(self.eCOUPx**2 + self.eCOUPy**2)
                    * np.sin(np.pi / 3 + fib + angular_position)
                )
            )
            * np.cos(4 * np.pi / 3 + angular_position)
        )

        self.mi_x = (
            (
                np.sqrt(
                    self.radius**2
                    + self.eCOUPx**2
                    + self.eCOUPy**2
                    + 2
                    * self.radius
                    * np.sqrt(self.eCOUPx**2 + self.eCOUPy**2)
                    * np.sin(fib + angular_position)
                )
                - self.radius
            )
            * np.sin(angular_position)
            + (
                np.sqrt(
                    self.radius**2
                    + self.eCOUPx**2
                    + self.eCOUPy**2
                    + 2
                    * self.radius
                    * np.sqrt(self.eCOUPx**2 + self.eCOUPy**2)
                    * np.cos(np.pi / 6 + fib + angular_position)
                )
                - self.radius
            )
            * np.sin(2 * np.pi / 3 + angular_position)
            + (
                self.radius
                - np.sqrt(
                    self.radius**2
                    + self.eCOUPx**2
                    + self.eCOUPy**2
                    - 2
                    * self.radius
                    * np.sqrt(self.eCOUPx**2 + self.eCOUPy**2)
                    * np.sin(np.pi / 3 + fib + angular_position)
                )
            )
            * np.sin(4 * np.pi / 3 + angular_position)
        )

        Fpy = self.kd * self.mi_y

        Fpx = self.kd * self.mi_x

        F_mis_p[0 + 6 * self.n1] = Fpx
        F_mis_p[1 + 6 * self.n1] = Fpy
        F_mis_p[5 + 6 * self.n1] = self.TD
        F_mis_p[0 + 6 * self.n2] = -Fpx
        F_mis_p[1 + 6 * self.n2] = -Fpy
        F_mis_p[5 + 6 * self.n2] = self.TL

        return F_mis_p

    def _angular(self, angular_position):
        """Reaction forces of angular misalignment.

        angular_position : float
                Angular position of the shaft.

        Returns
        -------
        F_mis_a : array
            Excitation force caused by the parallel misalignment on the entire system.
        """
        F_mis_a = np.zeros((self.ndof, len(angular_position)))

        Fay = (
            np.abs(
                self.Cte * np.sin(angular_position) * np.sin(self.misalignment_angle)
            )
            * np.sin(angular_position + np.pi)
            + np.abs(
                self.Cte
                * np.sin(angular_position + 2 * np.pi / 3)
                * np.sin(self.misalignment_angle)
            )
            * np.sin(angular_position + np.pi + 2 * np.pi / 3)
            + np.abs(
                self.Cte
                * np.sin(angular_position + 4 * np.pi / 3)
                * np.sin(self.misalignment_angle)
            )
            * np.sin(angular_position + np.pi + 4 * np.pi / 3)
        )

        Fax = (
            np.abs(
                self.Cte * np.sin(angular_position) * np.sin(self.misalignment_angle)
            )
            * np.cos(angular_position + np.pi)
            + np.abs(
                self.Cte
                * np.sin(angular_position + 2 * np.pi / 3)
                * np.sin(self.misalignment_angle)
            )
            * np.cos(angular_position + np.pi + 2 * np.pi / 3)
            + np.abs(
                self.Cte
                * np.sin(angular_position + 4 * np.pi / 3)
                * np.sin(self.misalignment_angle)
            )
            * np.cos(angular_position + np.pi + 4 * np.pi / 3)
        )

        F_mis_a[0 + 6 * self.n1] = Fax
        F_mis_a[1 + 6 * self.n1] = Fay
        F_mis_a[5 + 6 * self.n1] = self.TD
        F_mis_a[0 + 6 * self.n2] = -Fax
        F_mis_a[1 + 6 * self.n2] = -Fay
        F_mis_a[5 + 6 * self.n2] = self.TL

        return F_mis_a

    def _combined(self, angular_position):
        """Reaction forces of combined (parallel and angular) misalignment.

        angular_position : float
                Angular position of the shaft.

        Returns
        -------
        F_misalign : array
            Excitation force caused by the parallel misalignment on the entire system.
        """
        F_misalign = self._parallel(angular_position) + self._angular(angular_position)
        return F_misalign


class MisalignmentRigid(Fault):
    """A rigid coupling with parallel misalignment.

    Calculates the dynamic reaction force of hexangular rigid coupling
    induced by 6DOF's rotor parallel misalignment.

    Parameters
    ----------
    dt : float
        Time step.
    tI : float
        Initial time.
    tF : float
        Final time.
    eCOUP : float
        Parallel misalignment offset between driving rotor and driven rotor along X direction.
    TD : float
        Driving torque.
    TL : float
        Driven torque.
    n1 : float
        Node where the misalignment is ocurring.
    speed : float, pint.Quantity
        Operational speed of the machine. Default unit is rad/s.
    unbalance_magnitude : array
        Array with the unbalance magnitude. The unit is kg.m.
    unbalance_phase : array
        Array with the unbalance phase. The unit is rad.
    print_progress : bool
        Set it True, to print the time iterations and the total time spent.
        False by default.

    Returns
    -------
    A force to be applied on the shaft.

    References
    ----------

    .. [1] 'Al-Hussain, K. M., & Redmond, I. (2002). Dynamic response of two rotors connected by rigid mechanical coupling with parallel misalignment. Journal of Sound and vibration, 249(3), 483-498..

    Examples
    --------
    >>> from ross.probe import Probe
    >>> from ross.faults.misalignment import misalignment_rigid_example
    >>> probe1 = Probe(14, 0)
    >>> probe2 = Probe(22, 0)
    >>> response = misalignment_rigid_example()
    >>> results = response.run_time_response()
    >>> fig = response.plot_dfft(probe=[probe1, probe2], range_freq=[0, 100], yaxis_type="log")
    >>> # fig.show
    """

    @check_units
    def __init__(
        self,
        dt,
        tI,
        tF,
        eCOUP,
        TD,
        TL,
        n1,
        speed,
        unbalance_magnitude,
        unbalance_phase,
        print_progress=False,
    ):
        self.dt = dt
        self.tI = tI
        self.tF = tF
        self.eCOUP = eCOUP
        self.TD = TD
        self.TL = TL
        self.n1 = n1
        self.n2 = n1 + 1
        self.speed = speed
        self.speedI = speed
        self.speedF = speed
        self.unbalance_magnitude = unbalance_magnitude
        self.unbalance_phase = unbalance_phase
        self.DoF = np.arange((self.n1 * 6), (self.n2 * 6 + 6))
        self.print_progress = print_progress

        if len(self.unbalance_magnitude) != len(self.unbalance_phase):
            raise Exception(
                "The unbalance magnitude vector and phase must have the same size!"
            )

    def run(self, rotor):
        """Calculates the shaft angular position and the misalignment amount at X directions.

        Parameters
        ----------
        rotor : ross.Rotor Object
             6 DoF rotor model.

        """
        self.rotor = rotor
        self.n_disk = len(self.rotor.disk_elements)
        if self.n_disk != len(self.unbalance_magnitude):
            raise Exception("The number of discs and unbalances must agree!")

        self.ndof = rotor.ndof
        self.ndofd = np.zeros(len(self.rotor.disk_elements))

        for ii in range(self.n_disk):
            self.ndofd[ii] = (self.rotor.disk_elements[ii].n) * 6

        self.lambdat = 0.00001
        # Faxial = 0
        # TorqueI = 0
        # TorqueF = 0

        # pre-processing of auxilary variuables for the time integration
        self.sA = (
            self.speedI * np.exp(-self.lambdat * self.tF)
            - self.speedF * np.exp(-self.lambdat * self.tI)
        ) / (np.exp(-self.lambdat * self.tF) - np.exp(-self.lambdat * self.tI))
        self.sB = (self.speedF - self.speedI) / (
            np.exp(-self.lambdat * self.tF) - np.exp(-self.lambdat * self.tI)
        )

        # This code below here is used for acceleration and torque application to the rotor. As of
        # september/2020 it is unused, but might be implemented in future releases. These would be
        # run-up and run-down operations and variations of operating conditions.
        #
        # sAT = (
        #     TorqueI * np.exp(-lambdat * self.tF) - TorqueF * np.exp(-lambdat * self.tI)
        # ) / (np.exp(-lambdat * self.tF) - np.exp(-lambdat * self.tI))
        # sBT = (TorqueF - TorqueI) / (
        #     np.exp(-lambdat * self.tF) - np.exp(-lambdat * self.tI)
        # )
        #
        # self.SpeedV = sA + sB * np.exp(-lambdat * t)
        # self.TorqueV = sAT + sBT * np.exp(-lambdat * t)
        # self.AccelV = -lambdat * sB * np.exp(-lambdat * t)

        # Determining the modal matrix
        self.K = self.rotor.K(self.speed)
        self.C = self.rotor.C(self.speed)
        self.G = self.rotor.G()
        self.M = self.rotor.M(self.speed)
        self.Ksdt = self.rotor.Ksdt()

        _, ModMat = la.eigh(self.K, self.M)
        ModMat = ModMat[:, :12]
        self.ModMat = ModMat

        # Modal transformations
        self.Mmodal = ((ModMat.T).dot(self.M)).dot(ModMat)
        self.Cmodal = ((ModMat.T).dot(self.C)).dot(ModMat)
        self.Gmodal = ((ModMat.T).dot(self.G)).dot(ModMat)
        self.Kmodal = ((ModMat.T).dot(self.K)).dot(ModMat)
        self.Ksdtmodal = ((ModMat.T).dot(self.Ksdt)).dot(ModMat)

        self.angANG = -np.pi / 180

        self.kcoup_auxt = 1 / (
            self.K[5 + 6 * self.n1, 5 + 6 * self.n1]
            + self.K[5 + 6 * self.n2, 5 + 6 * self.n2]
        )

        self.kCOUP = (
            self.K[6 * self.n1, 6 * self.n1] * self.K[6 * self.n2, 6 * self.n2]
        ) / (self.K[6 * self.n1, 6 * self.n1] + self.K[6 * self.n2, 6 * self.n2])

        self.Kcoup_auxI = self.K[5 + 6 * self.n1, 5 + 6 * self.n1] / (
            self.K[5 + 6 * self.n1, 5 + 6 * self.n1]
            + self.K[5 + 6 * self.n2, 5 + 6 * self.n2]
        )

        self.Kcoup_auxF = self.K[5 + 6 * self.n2, 5 + 6 * self.n2] / (
            self.K[5 + 6 * self.n1, 5 + 6 * self.n1]
            + self.K[5 + 6 * self.n2, 5 + 6 * self.n2]
        )

        FFmis = np.zeros(self.ndof)

        # Omega = self.speedI * np.pi / 30

        y0 = np.zeros(24)
        t_eval = np.arange(self.tI, self.tF + self.dt, self.dt)
        T = t_eval

        self.angular_position = (
            self.sA * T
            - (self.sB / self.lambdat) * np.exp(-self.lambdat * T)
            + (self.sB / self.lambdat)
        )

        self.Omega = self.sA + self.sB * np.exp(-self.lambdat * T)
        self.AccelV = -self.lambdat * self.sB * np.exp(-self.lambdat * T)

        self.tetaUNB = np.zeros((len(self.unbalance_phase), len(self.angular_position)))
        unbx = np.zeros(len(self.angular_position))
        unby = np.zeros(len(self.angular_position))

        FFunb = np.zeros((self.ndof, len(t_eval)))

        for ii in range(self.n_disk):
            self.tetaUNB[ii, :] = (
                self.angular_position + self.unbalance_phase[ii] + np.pi / 2
            )

            unbx = self.unbalance_magnitude[ii] * (self.AccelV) * (
                np.cos(self.tetaUNB[ii, :])
            ) - self.unbalance_magnitude[ii] * ((self.Omega**2)) * (
                np.sin(self.tetaUNB[ii, :])
            )

            unby = -self.unbalance_magnitude[ii] * (self.AccelV) * (
                np.sin(self.tetaUNB[ii, :])
            ) - self.unbalance_magnitude[ii] * (self.Omega**2) * (
                np.cos(self.tetaUNB[ii, :])
            )

            FFunb[int(self.ndofd[ii]), :] += unbx
            FFunb[int(self.ndofd[ii] + 1), :] += unby

        self.Funbmodal = (self.ModMat.T).dot(FFunb)
        self.forces = np.zeros((self.ndof, len(t_eval)))

        self.inv_Mmodal = np.linalg.pinv(self.Mmodal)
        t1 = time.time()

        x = Integrator(
            self.tI,
            y0,
            self.tF,
            self.dt,
            self._equation_of_movement,
            self.print_progress,
        )
        x = x.rk45()
        t2 = time.time()
        if self.print_progress:
            print(f"Time spent: {t2-t1} s")

        self.displacement = x[:12, :]
        self.velocity = x[12:, :]
        self.time_vector = t_eval
        self.response = self.ModMat.dot(self.displacement)

    def _equation_of_movement(self, T, Y, i):
        """Calculates the displacement and velocity using state-space representation in the modal domain.

        Parameters
        ----------
        T : float
            Iteration time.
        Y : array
            Array of displacement and velocity, in the modal domain.
        i : int
            Iteration step.

        Returns
        -------
        new_Y :  array
            Array of the new displacement and velocity, in the modal domain.
        """
        positions = Y[:12]
        velocity = Y[12:]  # velocity ign space state

        positionsFis = self.ModMat.dot(positions)

        self.angANG = (
            self.Kcoup_auxI * self.angular_position[i]
            + self.Kcoup_auxF * self.angular_position[i]
            + self.kCOUP
            * self.kcoup_auxt
            * self.eCOUP
            * (
                -positionsFis[0 + 6 * self.n1] * np.sin(self.angANG)
                + positionsFis[0 + 6 * self.n2] * np.sin(self.angANG)
                + positionsFis[1 + 6 * self.n1] * np.cos(self.angANG)
                - positionsFis[1 + 6 * self.n2] * np.cos(self.angANG)
            )
        )

        Fmis, ft = self._parallel(positionsFis, self.angANG)
        self.forces[:, i] = ft
        ftmodal = (self.ModMat.T).dot(ft)

        # proper equation of movement to be integrated in time
        new_V_dot = (
            ftmodal
            + self.Funbmodal[:, i]
            - ((self.Cmodal + self.Gmodal * self.Omega[i])).dot(velocity)
            - ((self.Kmodal + self.Ksdtmodal * self.AccelV[i]).dot(positions))
        ).dot(self.inv_Mmodal)

        new_X_dot = velocity

        new_Y = np.zeros(24)
        new_Y[:12] = new_X_dot
        new_Y[12:] = new_V_dot

        return new_Y

    def _parallel(self, positions, fir):
        """Reaction forces of parallel misalignment.

        Returns
        -------
        Fmis : array
            Excitation force caused by the parallel misalignment on the node of application.
        FFmis : array
            Excitation force caused by the parallel misalignment on the entire system.
        """
        k0 = self.kCOUP
        delta1 = self.eCOUP

        betam = 0

        k_misalignbeta1 = np.array(
            [
                k0 * self.Kcoup_auxI * delta1 * np.sin(betam + fir),
                -k0 * self.Kcoup_auxI * delta1 * np.cos(betam + fir),
                0,
                0,
                0,
                0,
                k0 * self.Kcoup_auxF * delta1 * np.sin(betam + fir),
                -k0 * self.Kcoup_auxF * delta1 * np.cos(betam + fir),
                0,
                0,
                0,
                0,
            ]
        )

        K_mis_matrix = np.zeros((12, 12))
        K_mis_matrix[5, :] = k_misalignbeta1
        K_mis_matrix[11, :] = -k_misalignbeta1

        Force_kkmis = K_mis_matrix.dot(positions[self.DoF])

        F_misalign = np.array(
            [
                (-k0 * delta1 * np.cos(betam + fir) + k0 * delta1),
                -k0 * delta1 * np.sin(betam + fir),
                0,
                0,
                0,
                self.TD - self.TL,
                (k0 * delta1 * np.cos(betam + fir) - k0 * delta1),
                k0 * delta1 * np.sin(betam + fir),
                0,
                0,
                0,
                -(self.TD - self.TL),
            ]
        )

        Fmis = Force_kkmis + F_misalign
        FFmis = np.zeros(self.ndof)
        FFmis[self.DoF] = Fmis

        return Fmis, FFmis


def base_rotor_example():
    """Internal routine that create an example of a rotor, to be used in
    the associated misalignment problems as a prerequisite.

    This function returns an instance of a 6 DoF rotor, with a number of
    components attached. As this is not the focus of the example here, but
    only a requisite, see the example in "rotor assembly" for additional
    information on the rotor object.

    Returns
    -------
    rotor : ross.Rotor Object
        An instance of a flexible 6 DoF rotor object.

    Examples
    --------
    >>> rotor = base_rotor_example()
    >>> rotor.Ip
    0.015118294226367068
    """
    steel2 = ross.Material(name="Steel", rho=7850, E=2.17e11, G_s=81.2e9)
    #  Rotor with 6 DoFs, with internal damping, with 10 shaft elements, 2 disks and 2 bearings.
    i_d = 0
    o_d = 0.019
    n = 33

    # fmt: off
    L = np.array(
            [0  ,  25,  64, 104, 124, 143, 175, 207, 239, 271,
            303, 335, 345, 355, 380, 408, 436, 466, 496, 526,
            556, 586, 614, 647, 657, 667, 702, 737, 772, 807,
            842, 862, 881, 914]
            )/ 1000
    # fmt: on

    L = [L[i] - L[i - 1] for i in range(1, len(L))]

    shaft_elem = [
        ross.ShaftElement6DoF(
            material=steel2,
            L=l,
            idl=i_d,
            odl=o_d,
            idr=i_d,
            odr=o_d,
            alpha=8.0501,
            beta=1.0e-5,
            rotary_inertia=True,
            shear_effects=True,
        )
        for l in L
    ]

    Id = 0.003844540885417
    Ip = 0.007513248437500

    disk0 = ross.DiskElement6DoF(n=12, m=2.6375, Id=Id, Ip=Ip)
    disk1 = ross.DiskElement6DoF(n=24, m=2.6375, Id=Id, Ip=Ip)

    kxx1 = 4.40e5
    kyy1 = 4.6114e5
    kzz = 0
    cxx1 = 27.4
    cyy1 = 2.505
    czz = 0
    kxx2 = 9.50e5
    kyy2 = 1.09e8
    cxx2 = 50.4
    cyy2 = 100.4553

    bearing0 = ross.BearingElement6DoF(
        n=4, kxx=kxx1, kyy=kyy1, cxx=cxx1, cyy=cyy1, kzz=kzz, czz=czz
    )
    bearing1 = ross.BearingElement6DoF(
        n=31, kxx=kxx2, kyy=kyy2, cxx=cxx2, cyy=cyy2, kzz=kzz, czz=czz
    )

    rotor = ross.Rotor(shaft_elem, [disk0, disk1], [bearing0, bearing1])

    return rotor


def misalignment_flex_parallel_example():
    """Create an example of a flexible parallel misalignment fault.

    This function returns an instance of a flexible parallel misalignment
    fault. The purpose is to make available a simple model so that a
    doctest can be written using it.

    Returns
    -------
    misalignment : ross.MisalignmentFlex Object
        An instance of a flexible parallel misalignment model object.

    Examples
    --------
    >>> misalignment = misalignment_flex_parallel_example()
    >>> misalignment.speed
    125.66370614359172
    """

    rotor = base_rotor_example()

    misalignment = rotor.run_misalignment(
        coupling="flex",
        dt=0.0001,
        tI=0,
        tF=0.5,
        kd=40 * 10 ** (3),
        ks=38 * 10 ** (3),
        eCOUPx=2 * 10 ** (-4),
        eCOUPy=2 * 10 ** (-4),
        misalignment_angle=5 * np.pi / 180,
        TD=0,
        TL=0,
        n1=0,
        speed=Q_(1200, "RPM"),
        unbalance_magnitude=np.array([5e-4, 0]),
        unbalance_phase=np.array([-np.pi / 2, 0]),
        mis_type="parallel",
        print_progress=False,
    )

    return misalignment


def misalignment_flex_angular_example():
    """Create an example of a flexible angular misalignment fault.

    This function returns an instance of a flexible angular misalignment
    fault. The purpose is to make available a simple model so that a
    doctest can be written using it.

    Returns
    -------
    misalignment : ross.MisalignmentFlex Object
        An instance of a flexible Angular misalignment model object.

    Examples
    --------
    >>> misalignment = misalignment_flex_angular_example()
    >>> misalignment.speed
    125.66370614359172
    """

    rotor = base_rotor_example()

    misalignment = rotor.run_misalignment(
        coupling="flex",
        dt=0.0001,
        tI=0,
        tF=0.5,
        kd=40 * 10 ** (3),
        ks=38 * 10 ** (3),
        eCOUPx=2 * 10 ** (-4),
        eCOUPy=2 * 10 ** (-4),
        misalignment_angle=5 * np.pi / 180,
        TD=0,
        TL=0,
        n1=0,
        speed=Q_(1200, "RPM"),
        unbalance_magnitude=np.array([5e-4, 0]),
        unbalance_phase=np.array([-np.pi / 2, 0]),
        mis_type="angular",
        print_progress=False,
    )

    return misalignment


def misalignment_flex_combined_example():
    """Create an example of a flexible combined misalignment fault.

    This function returns an instance of a flexible combined misalignment
    fault. The purpose is to make available a simple model so that a
    doctest can be written using it.

    Returns
    -------
    misalignment : ross.MisalignmentFlex Object
        An instance of a flexible combined misalignment model object.

    Examples
    --------
    >>> misalignment = misalignment_flex_combined_example()
    >>> misalignment.speed
    125.66370614359172
    """

    rotor = base_rotor_example()

    misalignment = rotor.run_misalignment(
        coupling="flex",
        dt=0.0001,
        tI=0,
        tF=0.5,
        kd=40 * 10 ** (3),
        ks=38 * 10 ** (3),
        eCOUPx=2 * 10 ** (-4),
        eCOUPy=2 * 10 ** (-4),
        misalignment_angle=5 * np.pi / 180,
        TD=0,
        TL=0,
        n1=0,
        speed=Q_(1200, "RPM"),
        unbalance_magnitude=np.array([5e-4, 0]),
        unbalance_phase=np.array([-np.pi / 2, 0]),
        mis_type="combined",
        print_progress=False,
    )

    return misalignment


def misalignment_rigid_example():
    """Create an example of a rigid misalignment fault.

    This function returns an instance of a rigid misalignment
    fault. The purpose is to make available a simple model so that a
    doctest can be written using it.

    Returns
    -------
    misalignment : ross.MisalignmentRigid Object
        An instance of a rigid misalignment model object.

    Examples
    --------
    >>> misalignment = misalignment_rigid_example()
    >>> misalignment.speed
    125.66370614359172
    """

    rotor = base_rotor_example()

    misalignment = rotor.run_misalignment(
        coupling="rigid",
        dt=0.0001,
        tI=0,
        tF=0.5,
        eCOUP=2e-4,
        TD=0,
        TL=0,
        n1=0,
        speed=Q_(1200, "RPM"),
        unbalance_magnitude=np.array([5e-4, 0]),
        unbalance_phase=np.array([-np.pi / 2, 0]),
        print_progress=False,
    )

    return misalignment



import sys

import numpy as np
import scipy as sp

from ross.fluid_flow.fluid_flow_coefficients import find_equilibrium_position
from ross.fluid_flow.fluid_flow_geometry import (
    calculate_attitude_angle,
    calculate_eccentricity_ratio,
    calculate_rotor_load,
    external_radius_function,
    internal_radius_function,
    modified_sommerfeld_number,
)


class FluidFlow:
    r"""Generate dynamic coefficients for hydrodynamic bearings.

    This class calculate the pressure matrix and the stiffness and damping matrices
    of a fluid flow with the given parameters.

    It is supposed to be an attribute of a bearing element,
    but can work on its on to provide graphics for the user.

    The complete FluidFlow theory can be found at :cite:`mota2020`

    Parameters
    ----------
    Grid related
    ^^^^^^^^^^^^
    Describes the discretization of the problem
    nz: int
        Number of points along the Z direction (direction of flow).
    ntheta: int
        Number of points along the direction theta. NOTE: ntheta must be odd.
    length: float
        Length in the Z direction (m).

    Operation conditions
    ^^^^^^^^^^^^^^^^^^^^
    Describes the operation conditions.
    omega: float
        Rotation of the rotor (rad/s).
    p_in: float
        Input Pressure (Pa).
    p_out: float
        Output Pressure (Pa).
    load: float
        Load applied to the rotor (N).
    omegap: float
        Frequency of the rotor (rad/s).

    Geometric data of the problem
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    Describes the geometric data of the problem.
    radius_rotor: float
        Rotor radius (m).
    radius_stator: float
        Stator Radius (m).
    eccentricity: float, optional
        Eccentricity (m) is the euclidean distance between rotor and stator centers.
        The center of the stator is in position (0,0).
    attitude_angle: float, optional
        Attitude angle. Angle between the load line and the eccentricity (rad).
    bearing_type: str
        type of structure. 'short_bearing': short; 'long_bearing': long;
        The default is None. In this case it is automatically calculated
        following the parameter:
        'medium_size': in between short and long.
        if length/diameter <= 1/4 it is short.
        if length/diameter > 8 it is long.
    bearing_type: str, optional
        type of structure. 'short_bearing': short; 'long_bearing': long;
        The default is None. In this case it is automatically calculated
        following the parameter:
        'medium_size': in between short and long.
        if length/diameter <= 1/4 it is short.
        if length/diameter > 8 it is long.
    shape_geometry: str, optional
        Determines the type of bearing geometry.
        'cylindrical': cylindrical bearing; 'eliptical': eliptical bearing;
        'wear': journal bearing wear.
        The default is 'cylindrical'.
    preload: float
        Ellipticity ratio. The value must be between 0 and 1. If preload = 0
        the bearing becomes cylindrical. Not used in cylindrical bearings.
        The default is 0.4.
    displacement : float, optional
        Angular displacement of the bearing wear in relation to the vertical axis.
        Only necessary if shape_geometry is wear.
    max_depth: float
        The maximum wear depth. Only necessary if shape_geometry is wear..

    Fluid characteristics
    ^^^^^^^^^^^^^^^^^^^^^
    Describes the fluid characteristics.
    viscosity: float
        Viscosity (Pa.s).
    density: float
        Fluid density(Kg/m^3).

    User commands
    ^^^^^^^^^^^^^
    Commands that can be passed as arguments.
    immediately_calculate_pressure_matrix_numerically: bool, optional
        If set True, calculates the pressure matrix numerically immediately.

    Returns
    -------
    An object containing the fluid flow and its data.

    Attributes
    ----------
    ltheta: float
        Length in the theta direction (rad).
    dz: float
        Range size in the Z direction.
    dtheta: float
        Range size in the theta direction.
    ntotal: int
        Number of nodes in the grid., ntheta, n_interv_z, n_interv_theta,
    n_interv_z: int
        Number of intervals on Z.
    n_interv_theta: int
      Number of intervals on theta.
    p_mat_analytical : array of shape (nz, ntheta)
        The analytical pressure matrix.
    p_mat_numerical : array of shape (nz, ntheta)
        The numerical pressure matrix.
    xi: float
        Eccentricity (m) (distance between rotor and stator centers) on the x-axis.
        It is the position of the center of the rotor.
        The center of the stator is in position (0,0).
    yi: float
        Eccentricity (m) (distance between rotor and stator centers) on the y-axis.
        It is the position of the center of the rotor.
        The center of the stator is in position (0,0).
    re : array of shape (nz, ntheta)
        The external radius in each position of the grid.
    ri : array of shape (nz, ntheta)
        The internal radius in each position of the grid.
    xre : array of shape (nz, ntheta)
        x value of the external radius.
    xri : array of shape (nz, ntheta)
        x value of the internal radius.
    yre : array of shape (nz, ntheta)
        y value of the external radius.
    yri : array of shape (nz, ntheta)
        y value of the internal radius.
    z_list : array of shape (1, nz)
        z along the object. It goes from 0 to lb.
    gama : array of shape (nz, ntheta)
        Points along the object in the tangential direction.
        It ranges from 0 to 2 pi, starting at the largest spacing between rotor and stator.
    t : float
        Time.
    xp : float
        Perturbation along x.
    yp : float
        Perturbation along y.
    eccentricity : float
        distance between the center of the rotor and the stator.
    radial_clearance: float
        Difference between both stator and rotor radius, regardless of eccentricity.
    eccentricity_ratio: float
        eccentricity/radial_clearance
    characteristic_speed: float
        Characteristic fluid speeds.
        In journal bearings, characteristic_speed = omega * radius_rotor
    bearing_type: str
        type of structure. 'short_bearing': short; 'long_bearing': long;
        'medium_size': in between short and long.
        if length/diameter <= 1/4 it is short.
        if length/diameter > 8 it is long.
    analytical_pressure_matrix_available: bool
        True if analytically calculated pressure matrix is available.
    numerical_pressure_matrix_available: bool
        True if numerically calculated pressure matrix is available.

    References
    ----------
    .. bibliography:: ../../../docs/refs.bib

    Examples
    --------
    >>> from ross.fluid_flow import fluid_flow as flow
    >>> from ross.fluid_flow.fluid_flow_graphics import plot_pressure_theta
    >>> import numpy as np
    >>> nz = 8
    >>> ntheta = 64
    >>> length = 0.01
    >>> omega = 100.*2*np.pi/60
    >>> p_in = 0.
    >>> p_out = 0.
    >>> radius_rotor = 0.08
    >>> radius_stator = 0.1
    >>> viscosity = 0.015
    >>> density = 860.
    >>> eccentricity = 0.001
    >>> attitude_angle = np.pi
    >>> my_fluid_flow = flow.FluidFlow(nz, ntheta, length,
    ...                                omega, p_in, p_out, radius_rotor,
    ...                                radius_stator, viscosity, density,
    ...                                attitude_angle=attitude_angle, eccentricity=eccentricity,
    ...                                immediately_calculate_pressure_matrix_numerically=False)
    >>> my_fluid_flow.calculate_pressure_matrix_analytical() # doctest: +ELLIPSIS
    array([[...
    >>> my_fluid_flow.calculate_pressure_matrix_numerical() # doctest: +ELLIPSIS
    array([[...
    >>> # to show the plots you can use:
    >>> # my_fluid_flow.plot_eccentricity().show()
    >>> # my_fluid_flow.plot_pressure_theta(z=int(nz/2)).show()
    >>> fig = plot_pressure_theta(my_fluid_flow, z=int(nz/2)) # doctest: +ELLIPSIS
    >>> # fig.show() to display the figure.
    """

    def __init__(
        self,
        nz,
        ntheta,
        length,
        omega,
        p_in,
        p_out,
        radius_rotor,
        radius_stator,
        viscosity,
        density,
        attitude_angle=None,
        eccentricity=None,
        load=None,
        omegap=None,
        immediately_calculate_pressure_matrix_numerically=True,
        bearing_type=None,
        shape_geometry="cylindrical",
        preload=0.4,
        displacement=0,
        max_depth=None,
    ):
        self.nz = nz
        self.ntheta = ntheta
        self.n_interv_z = nz - 1
        self.n_interv_theta = ntheta - 1
        self.length = length
        self.ltheta = 2.0 * np.pi
        self.dz = length / self.n_interv_z
        self.dtheta = self.ltheta / self.n_interv_theta
        self.ntotal = self.nz * self.ntheta
        self.omega = omega
        self.p_in = p_in
        self.p_out = p_out
        self.radius_rotor = radius_rotor
        self.radius_stator = radius_stator
        self.viscosity = viscosity
        self.density = density
        self.characteristic_speed = self.omega * self.radius_rotor
        self.radial_clearance = self.radius_stator - self.radius_rotor
        self.bearing_type = bearing_type
        if bearing_type is None:
            if self.length / (2 * self.radius_stator) <= 1 / 4:
                self.bearing_type = "short_bearing"
            elif self.length / (2 * self.radius_stator) > 4:
                self.bearing_type = "long_bearing"
            else:
                self.bearing_type = "medium_size"
        self.shape_geometry = shape_geometry
        self.preload = preload
        self.displacement = displacement
        self.max_depth = max_depth
        self.eccentricity = eccentricity
        self.attitude_angle = attitude_angle
        self.eccentricity_ratio = None
        self.load = load

        self.omegap = omegap
        if self.omegap is None:
            self.omegap = self.omega
        else:
            self.omegap = omegap
        self.z_list = np.zeros(self.nz)
        self.re = np.zeros([self.nz, self.ntheta])
        self.ri = np.zeros([self.nz, self.ntheta])
        self.xre = np.zeros([self.nz, self.ntheta])
        self.xri = np.zeros([self.nz, self.ntheta])
        self.yre = np.zeros([self.nz, self.ntheta])
        self.yri = np.zeros([self.nz, self.ntheta])
        self.gama = np.zeros([self.nz, self.ntheta])
        self.t = 0
        self.xp = 0
        self.yp = 0
        if (
            self.bearing_type == "short_bearing"
            and self.shape_geometry == "cylindrical"
        ):
            if self.eccentricity is None and load is not None:
                modified_s = modified_sommerfeld_number(
                    self.radius_stator,
                    self.omega,
                    self.viscosity,
                    self.length,
                    self.load,
                    self.radial_clearance,
                )
                self.eccentricity_ratio = calculate_eccentricity_ratio(modified_s)
                self.eccentricity = (
                    calculate_eccentricity_ratio(modified_s) * self.radial_clearance
                )
                if attitude_angle is None:
                    self.attitude_angle = calculate_attitude_angle(
                        self.eccentricity_ratio
                    )
            elif self.eccentricity is not None and load is not None:
                modified_s = modified_sommerfeld_number(
                    self.radius_stator,
                    self.omega,
                    self.viscosity,
                    self.length,
                    self.load,
                    self.radial_clearance,
                )
                self.eccentricity_ratio = calculate_eccentricity_ratio(modified_s)
                if attitude_angle is None:
                    self.attitude_angle = calculate_attitude_angle(
                        self.eccentricity_ratio
                    )
            elif eccentricity is not None and load is None:
                self.eccentricity_ratio = self.eccentricity / self.radial_clearance
                self.load = calculate_rotor_load(
                    self.radius_stator,
                    self.omega,
                    self.viscosity,
                    self.length,
                    self.radial_clearance,
                    self.eccentricity_ratio,
                )
                if attitude_angle is None:
                    self.attitude_angle = calculate_attitude_angle(
                        self.eccentricity_ratio
                    )
            else:
                sys.exit("Either load or eccentricity must be given.")

            self.xi = self.eccentricity * np.cos(3 * np.pi / 2 + self.attitude_angle)
            self.yi = self.eccentricity * np.sin(3 * np.pi / 2 + self.attitude_angle)
            self.geometry_description()

        else:
            if load is not None:
                find_equilibrium_position(self)
                if eccentricity is not None:
                    self.eccentricity = eccentricity
                if attitude_angle is not None:
                    self.attitude_angle = attitude_angle
            else:
                if attitude_angle is None:
                    sys.exit("Attitude angle or load must be given.")
                if eccentricity is None:
                    sys.exit("Eccentricity or load must be given.")
            self.geometry_description()
            self.eccentricity_ratio = self.eccentricity / self.radial_clearance
            self.xi = self.eccentricity * np.cos(3 * np.pi / 2 + self.attitude_angle)
            self.yi = self.eccentricity * np.sin(3 * np.pi / 2 + self.attitude_angle)

        self.p_mat_analytical = np.zeros([self.nz, self.ntheta])
        self.p_mat_numerical = np.zeros([self.nz, self.ntheta])
        self.analytical_pressure_matrix_available = False
        self.numerical_pressure_matrix_available = False

        if immediately_calculate_pressure_matrix_numerically:
            self.calculate_pressure_matrix_numerical()

    def calculate_pressure_matrix_analytical(self, method=0, force_type=None):
        """This function calculates the pressure matrix analytically
        for the cylindrical bearing.
        Parameters
        ----------
        method: int
            Determines the analytical method to be used, when more than one is available.
            In case of a short bearing:
                0: based on the book Tribology Series vol. 33, by Frene et al., chapter 5.
                1: based on the chapter Linear and Nonlinear Rotordynamics, by Ishida and
                Yamamoto, from the book Flow-Induced Vibrations.
            In case of a long bearing:
                0: based on the Fundamentals of Fluid Flow Lubrification, by Hamrock, chapter 10.
        force_type: str
            If set, calculates the pressure matrix analytically considering the chosen type: 'short' or 'long'.
        Returns
        -------
        p_mat_analytical: matrix of float
            Pressure matrix of size (nz x ntheta)
        Examples
        --------
        >>> my_fluid_flow = fluid_flow_example()
        >>> my_fluid_flow.calculate_pressure_matrix_analytical() # doctest: +ELLIPSIS
        array([[...
        """
        if self.bearing_type == "short_bearing" or force_type == "short":
            if method == 0:
                for i in range(0, self.nz):
                    for j in range(0, self.ntheta):
                        # fmt: off
                        self.p_mat_analytical[i, j] = (
                                ((-3 * self.viscosity * self.omega) / self.radial_clearance ** 2) *
                                ((i * self.dz - (self.length / 2)) ** 2 - (self.length ** 2) / 4) *
                                (self.eccentricity_ratio * np.sin(j * self.dtheta)) /
                                (1 + self.eccentricity_ratio * np.cos(j * self.dtheta)) ** 3)
                        # fmt: on
                        if self.p_mat_analytical[i, j] < 0:
                            self.p_mat_analytical[i, j] = 0
            elif method == 1:
                for i in range(0, self.nz):
                    for j in range(0, self.ntheta):
                        # fmt: off
                        self.p_mat_analytical[i, j] = (3 * self.viscosity / ((self.radial_clearance ** 2) *
                                                                             (1. + self.eccentricity_ratio * np.cos(
                                                                                 j * self.dtheta)) ** 3)) * \
                                                      (-self.eccentricity_ratio * self.omega * np.sin(
                                                          j * self.dtheta)) * \
                                                      (((i * self.dz - (self.length / 2)) ** 2) - (
                                                              self.length ** 2) / 4)
                        # fmt: on
                        if self.p_mat_analytical[i, j] < 0:
                            self.p_mat_analytical[i, j] = 0
        elif self.bearing_type == "long_bearing" or force_type == "long":
            if method == 0:
                for i in range(0, self.nz):
                    for j in range(0, self.ntheta):
                        self.p_mat_analytical[i, j] = (
                            6
                            * self.viscosity
                            * self.omega
                            * (self.radius_rotor / self.radial_clearance) ** 2
                            * self.eccentricity_ratio
                            * np.sin(self.dtheta * j)
                            * (2 + self.eccentricity_ratio * np.cos(self.dtheta * j))
                        ) / (
                            (2 + self.eccentricity_ratio**2)
                            * (1 + self.eccentricity_ratio * np.cos(self.dtheta * j))
                            ** 2
                        ) + self.p_in
                        if self.p_mat_analytical[i, j] < 0:
                            self.p_mat_analytical[i, j] = 0
        elif self.bearing_type == "medium_size" or self.shape_geometry != "cylindrical":
            raise ValueError(
                "The pressure matrix can only be calculated analytically for short or long cylindrical "
                "bearings. For cylindrical bearings: Try calling calculate_pressure_matrix_numerical "
                "or setting force_type to either 'short' or 'long' in calculate_pressure_matrix_analytical. "
                "For other geometries: Try calling calculate_pressure_matrix_numerical"
            )
        self.analytical_pressure_matrix_available = True
        return self.p_mat_analytical

    def geometry_description(self):
        """This function calculates the geometry description.
        It is executed when the class is instantiated.
        Examples
        --------
        >>> my_fluid_flow = fluid_flow_example()
        >>> my_fluid_flow.geometry_description()
        """
        if self.shape_geometry == "cylindrical":
            start = (np.pi / 2) + self.attitude_angle
        else:
            start = 0

        for i in range(0, self.nz):
            zno = i * self.dz
            self.z_list[i] = zno
            for j in range(0, self.ntheta):
                # fmt: off
                self.gama[i, j] = j * self.dtheta + start
                [radius_external, self.xre[i, j], self.yre[i, j]] = \
                    external_radius_function(self.gama[i, j], self.radius_stator, self.radius_rotor,
                                             shape=self.shape_geometry, preload=self.preload,
                                             displacement=self.displacement, max_depth=self.max_depth)
                [radius_internal, self.xri[i, j], self.yri[i, j]] = \
                    internal_radius_function(self.gama[i, j], self.attitude_angle, self.radius_rotor,
                                             self.eccentricity)
                self.re[i, j] = radius_external
                self.ri[i, j] = radius_internal
                # fmt: on

    def calculate_coefficients(self, direction=None):
        """This function calculates the constants that form the Poisson equation
        of the discrete pressure (central differences in the second
        derivatives).
        Parameters
        ----------
        direction: str
            If defined, it calculates the model based on the disturbance in the chosen direction: 'x' or 'y'.
        Returns
        --------
        c1, c2, c0w: matrix of float
            Constants that form the Poisson equation.
        >>> my_fluid_flow = fluid_flow_example()
        >>> my_fluid_flow.calculate_coefficients()# doctest: +ELLIPSIS
        (array([[...
        """
        c1 = np.zeros([self.nz, self.ntheta])
        c2 = np.zeros([self.nz, self.ntheta])
        c0w = np.zeros([self.nz, self.ntheta])

        for i in range(0, self.nz):
            eccentricity_error = False
            for j in range(0, self.ntheta):
                # fmt: off

                w = self.omega * self.radius_rotor

                k = (self.re[i, j] ** 2 * (np.log(self.re[i, j]) - 1 / 2) - self.ri[i, j] ** 2 *
                     (np.log(self.ri[i, j]) - 1 / 2)) / (self.ri[i, j] ** 2 - self.re[i, j] ** 2)

                c1[i, j] = (1 / (4 * self.viscosity)) * ((self.re[i, j] ** 2 * np.log(self.re[i, j]) -
                                                               self.ri[i, j] ** 2 * np.log(self.ri[i, j]) +
                                                               (self.re[i, j] ** 2 - self.ri[i, j] ** 2) *
                                                               (k - 1)) - 2 * self.re[i, j] ** 2 * (
                                                                      (np.log(self.re[i, j]) + k - 1 / 2) * np.log(
                                                                       self.re[i, j] / self.ri[i, j])))

                c2[i, j] = (- self.ri[i, j] ** 2) / (8 * self.viscosity) * \
                                ((self.re[i, j] ** 2 - self.ri[i, j] ** 2 -
                                  (self.re[i, j] ** 4 - self.ri[i, j] ** 4) /
                                  (2 * self.ri[i, j] ** 2)) +
                                 ((self.re[i, j] ** 2 - self.ri[i, j] ** 2) /
                                  (self.ri[i, j] ** 2 *
                                   np.log(self.re[i, j] / self.ri[i, j]))) *
                                 (self.re[i, j] ** 2 * np.log(self.re[i, j] / self.ri[i, j]) -
                                  (self.re[i, j] ** 2 - self.ri[i, j] ** 2) / 2))

                c0w[i, j] = (- w * self.ri[i, j] *
                                  (np.log(self.re[i, j] / self.ri[i, j]) *
                                   (1 + (self.ri[i, j] ** 2) / (self.re[i, j] ** 2 - self.ri[i, j] ** 2)) - 1 / 2))
                if direction == "x":
                    a = self.omegap * self.xp * np.cos(self.omegap * self.t)
                    c0w[i, j] += self.ri[i, j] * a * np.sin(self.gama[i, j])
                elif direction == "y":
                    b = self.omegap * self.yp * np.cos(self.omegap * self.t)
                    c0w[i, j] -= self.ri[i, j] * b * np.cos(self.gama[i, j])
                else:
                    c0w[i, j] += 0
                # fmt: on
                if not eccentricity_error:
                    if abs(self.xri[i, j]) > abs(self.xre[i, j]) or abs(
                        self.yri[i, j]
                    ) > abs(self.yre[i, j]):
                        eccentricity_error = True
            if eccentricity_error:
                raise ValueError(
                    "Error: The given parameters create a rotor that is not inside the stator. "
                    "Check parameters and fix accordingly."
                )
        return c1, c2, c0w

    def mounting_matrix(self, c1, c2, c0w):
        """This function assembles the matrix M and the independent vector f.
        Parameters
        ----------
        c1, c2, c0w: matrix of float
            Constants that form the Poisson equation.
        Returns
        --------
        M: matrix of float
            Matrix composed of coefficients that multiply the pressures at each point in the discrete domain.
        f: array of float
            Pressure independent terms.
        Examples
        --------
        >>> my_fluid_flow = fluid_flow_example()
        >>> c1, c2, c0w = my_fluid_flow.calculate_coefficients()
        >>> my_fluid_flow.mounting_matrix(c1, c2, c0w)# doctest: +ELLIPSIS
        (array([[...
        """
        # fmt: off
        M = np.zeros([self.ntotal, self.ntotal])
        f = np.zeros([self.ntotal, 1])
        count = 0
        for x in range(self.ntheta):
            M[count, count] = 1
            f[count, 0] = self.p_in
            count = count + self.nz - 1
            M[count, count] = 1
            f[count, 0] = self.p_out
            count = count + 1
        count = 0
        for x in range(self.nz - 2):
            M[self.ntotal - self.nz + 1 + count, 1 + count] = 1
            M[self.ntotal - self.nz + 1 + count, self.ntotal - self.nz + 1 + count] = -1
            count = count + 1
        count = 1
        j = 0
        for i in range(1, self.nz - 1):
            a = (1 / self.dtheta ** 2) * (c1[i, self.ntheta - 1])
            M[count, self.ntotal - 2 * self.nz + count] = a
            b = (1 / self.dz ** 2) * (c2[i - 1, j])
            M[count, count - 1] = b
            c = -((1 / self.dtheta ** 2) * ((c1[i, j]) + c1[i, self.ntheta - 1])
                  + (1 / self.dz ** 2) * (c2[i, j] + c2[i - 1, j]))
            M[count, count] = c
            d = (1 / self.dz ** 2) * (c2[i, j])
            M[count, count + 1] = d
            e = (1 / self.dtheta ** 2) * (c1[i, j])
            M[count, count + self.nz] = e
            count = count + 1
        count = self.nz + 1
        for j in range(1, self.ntheta - 1):
            for i in range(1, self.nz - 1):
                a = (1 / self.dtheta ** 2) * (c1[i, j - 1])
                M[count, count - self.nz] = a
                b = (1 / self.dz ** 2) * (c2[i - 1, j])
                M[count, count - 1] = b
                c = -((1 / self.dtheta ** 2) * ((c1[i, j]) + c1[i, j - 1])
                      + (1 / self.dz ** 2) * (c2[i, j] + c2[i - 1, j]))
                M[count, count] = c
                d = (1 / self.dz ** 2) * (c2[i, j])
                M[count, count + 1] = d
                e = (1 / self.dtheta ** 2) * (c1[i, j])
                M[count, count + self.nz] = e
                count = count + 1
            count = count + 2
        count = 1
        for j in range(self.ntheta - 1):
            for i in range(1, self.nz - 1):
                if j == 0:
                    f[count, 0] = (c0w[i, j] - c0w[i, self.ntheta - 1]) / self.dtheta
                else:
                    f[count, 0] = (c0w[i, j] - c0w[i, j - 1]) / self.dtheta
                count = count + 1
            count = count + 2
        # fmt: on
        return M, f

    def resolves_matrix(self, M, f):
        """This function resolves the linear system [M]{P} = {f}.
         Parameters
        ----------
        M: matrix of float
            Matrix composed of coefficients that multiply the pressures at each point in the discrete domain.
        f: array of float
            Pressure independent terms.
        Returns
        --------
        P: array of floats
            Pressure. Unknowns of the finite difference system.
        Examples
        --------
        >>> my_fluid_flow = fluid_flow_example()
        >>> c1, c2, c0w = my_fluid_flow.calculate_coefficients()
        >>> M, f = my_fluid_flow.mounting_matrix(c1, c2, c0w)
        >>> my_fluid_flow.resolves_matrix(M, f)# doctest: +ELLIPSIS
        array([[...
        """
        sparse_matrix = sp.sparse.csc_matrix(M)
        P = sp.sparse.linalg.spsolve(sparse_matrix, f)
        P.shape = (P.size, 1)
        return P

    def calculate_pressure_matrix_numerical(self, direction=None):
        """This function calculates the pressure matrix numerically.
        Parameters
        ----------
        direction: str
            If defined, it calculates the model based on the disturbance in the chosen direction: 'x' or 'y'.
        Returns
        -------
        p_mat_numerical: matrix of float
            Pressure matrix of size (nz x ntheta)
        Examples
        --------
        >>> my_fluid_flow = fluid_flow_example()
        >>> my_fluid_flow.calculate_pressure_matrix_numerical() # doctest: +ELLIPSIS
        array([[...
        """
        if direction == "x":
            c1, c2, c0w = self.calculate_coefficients(direction="x")
        elif direction == "y":
            c1, c2, c0w = self.calculate_coefficients(direction="y")
        else:
            c1, c2, c0w = self.calculate_coefficients()
        M, f = self.mounting_matrix(c1, c2, c0w)
        P = self.resolves_matrix(M, f)
        self.p_mat_numerical = np.clip(
            P.reshape((self.ntheta, self.nz)), a_min=0, a_max=None
        ).T
        self.numerical_pressure_matrix_available = True
        return self.p_mat_numerical


def fluid_flow_example():
    """This function returns an instance of a simple fluid flow.
    The purpose is to make available a simple short-bearing model
    so that doctest can be written using it.

    Parameters
    ----------

    Returns
    -------
    An instance of a fluid flow object.

    Examples
    --------
    >>> my_fluid_flow = fluid_flow_example()
    >>> my_fluid_flow.eccentricity
    0.0001
    """
    my_pressure_matrix = FluidFlow(
        nz=8,
        ntheta=32,
        length=0.04,
        omega=100.0 * 2 * np.pi / 60,
        p_in=0.0,
        p_out=0.0,
        radius_rotor=0.2,
        radius_stator=0.2002,
        viscosity=0.015,
        density=860.0,
        eccentricity=0.0001,
        attitude_angle=np.pi / 4,
        immediately_calculate_pressure_matrix_numerically=False,
    )
    return my_pressure_matrix


def fluid_flow_example2():
    """This function returns a different instance of a simple fluid flow.
    The purpose is to make available a simple medium-bearing model
    so that doctest can be written using it.

    Parameters
    ----------

    Returns
    -------
    An instance of a fluid flow object.

    Examples
    --------
    >>> my_fluid_flow = fluid_flow_example2()
    >>> my_fluid_flow.load
    525
    """
    nz = 8
    ntheta = 16
    length = 0.03
    omega = 157.1
    p_in = 0.0
    p_out = 0.0
    radius_rotor = 0.0499
    radius_stator = 0.05
    load = 525
    visc = 0.1
    rho = 860.0
    return FluidFlow(
        nz,
        ntheta,
        length,
        omega,
        p_in,
        p_out,
        radius_rotor,
        radius_stator,
        visc,
        rho,
        load=load,
        immediately_calculate_pressure_matrix_numerically=True,
    )


def fluid_flow_example3():
    """This function returns a different instance of a simple fluid flow.
    The purpose is to make available a simple eliptical-bearing model
    so that doctest can be written using it.

    Parameters
    ----------

    Returns
    -------
    An instance of a fluid flow object.

    Examples
    --------
    >>> my_fluid_flow = fluid_flow_example3()
    >>> my_fluid_flow.load
    100
    """
    nz = 8
    ntheta = 32
    omega = 2500 * np.pi / 30.0
    p_in = 0.0
    p_out = 0.0
    radius_stator = (3 * 10 ** (-2)) / 2
    cr = 9 * 10 ** (-5)
    m = 0.4
    radius_rotor = radius_stator - cr
    length = 2 * 10 ** (-2)
    load = 100
    viscosity = 5.449 * 10 ** (-2)
    density = 881
    return FluidFlow(
        nz,
        ntheta,
        length,
        omega,
        p_in,
        p_out,
        radius_rotor,
        radius_stator,
        viscosity,
        density,
        load=load,
        shape_geometry="eliptical",
        preload=m,
    )


def fluid_flow_example4():
    """This function returns a different instance of a simple fluid flow.
    The purpose is to make available a simple wear-bearing model
    so that doctest can be written using it.

    Parameters
    ----------

    Returns
    -------
    An instance of a fluid flow object.

    Examples
    --------
    >>> my_fluid_flow = fluid_flow_example4()
    >>> my_fluid_flow.load
    18.9
    """
    nz = 8
    ntheta = 32
    omega = 1000 * np.pi / 30.0
    p_in = 0.0
    p_out = 0.0
    radius_stator = (30 / 2) * 10 ** (-3)
    cr = 90 * 10 ** (-6)
    radius_rotor = radius_stator - cr
    length = 20 * 10 ** (-3)
    load = 18.9
    viscosity = 1.044 / 10.0
    density = 881
    max_depth = 50 * 10 ** (-6)
    y = 10 * np.pi / 180
    return FluidFlow(
        nz,
        ntheta,
        length,
        omega,
        p_in,
        p_out,
        radius_rotor,
        radius_stator,
        viscosity,
        density,
        load=load,
        shape_geometry="wear",
        max_depth=max_depth,
        displacement=y,
    )



"""Disk element module for STOCHASTIC ROSS.

This module creates an instance of random disk element for stochastic
analysis.
"""

import numpy as np

from ross.disk_element import DiskElement
from ross.stochastic.st_materials import ST_Material
from ross.stochastic.st_results_elements import plot_histogram
from ross.units import check_units

__all__ = ["ST_DiskElement", "st_disk_example"]


class ST_DiskElement:
    """Random disk element.

    Creates an object containing a list with random instances of DiskElement.

    Parameters
    ----------
    n: int
        Node in which the disk will be inserted.
    m : float, list
        Mass of the disk element.
        Input a list to make it random.
    Id : float, list
        Diametral moment of inertia.
        Input a list to make it random.
    Ip : float, list
        Polar moment of inertia
        Input a list to make it random.
    tag : str, optional
        A tag to name the element
        Default is None
    color : str, optional
        A color to be used when the element is represented.
        Default is "Firebrick".
    is_random : list
        List of the object attributes to become random.
        Possibilities:
            ["m", "Id", "Ip"]

    Example
    -------
    >>> import numpy as np
    >>> import ross.stochastic as srs
    >>> elms = srs.ST_DiskElement(n=1,
    ...                           m=30.0,
    ...                           Id=np.random.uniform(0.20, 0.40, 5),
    ...                           Ip=np.random.uniform(0.15, 0.25, 5),
    ...                           is_random=["Id", "Ip"],
    ...                           )
    >>> len(list(iter(elms)))
    5
    """

    @check_units
    def __init__(
        self,
        n,
        m,
        Id,
        Ip,
        tag=None,
        color="Firebrick",
        is_random=None,
    ):
        attribute_dict = dict(n=n, m=m, Id=Id, Ip=Ip, tag=tag, color=color)

        self.is_random = is_random
        self.attribute_dict = attribute_dict

    def __iter__(self):
        """Return an iterator for the container.

        Returns
        -------
        An iterator over random disk elements.

        Examples
        --------
        >>> import ross.stochastic as srs
        >>> elm = srs.st_disk_example()
        >>> len(list(iter(elm)))
        2
        """
        return iter(self.random_var(self.is_random, self.attribute_dict))

    def __getitem__(self, key):
        """Return the value for a given key from attribute_dict.

        Parameters
        ----------
        key : str
            A class parameter as string.

        Raises
        ------
        KeyError
            Raises an error if the parameter doesn't belong to the class.

        Returns
        -------
        Return the value for the given key.

        Example
        -------
        >>> import numpy as np
        >>> import ross.stochastic as srs
        >>> elms = srs.ST_DiskElement(n=1,
        ...                           m=30.0,
        ...                           Id=np.random.uniform(0.20, 0.40, 5),
        ...                           Ip=np.random.uniform(0.15, 0.25, 5),
        ...                           is_random=["Id", "Ip"],
        ...                           )
        >>> elms["m"]
        30.0
        """
        if key not in self.attribute_dict.keys():
            raise KeyError("Object does not have parameter: {}.".format(key))

        return self.attribute_dict[key]

    def __setitem__(self, key, value):
        """Set new parameter values for the object.

        Function to change a parameter value.
        It's not allowed to add new parameters to the object.

        Parameters
        ----------
        key : str
            A class parameter as string.
        value : The corresponding value for the attrbiute_dict's key.
            ***check the correct type for each key in ST_DiskElement
            docstring.

        Raises
        ------
        KeyError
            Raises an error if the parameter doesn't belong to the class.

        Example
        -------
        >>> import numpy as np
        >>> import ross.stochastic as srs
        >>> elms = srs.ST_DiskElement(n=1,
        ...                           m=30.0,
        ...                           Id=np.random.uniform(0.20, 0.40, 5),
        ...                           Ip=np.random.uniform(0.15, 0.25, 5),
        ...                           is_random=["Id", "Ip"],
        ...                           )
        >>> elms["Id"] = np.linspace(0.1, 0.3, 5)
        >>> elms["Id"]
        array([0.1 , 0.15, 0.2 , 0.25, 0.3 ])
        """
        if key not in self.attribute_dict.keys():
            raise KeyError("Object does not have parameter: {}.".format(key))
        self.attribute_dict[key] = value

    def random_var(self, is_random, *args):
        """Generate a list of objects as random attributes.

        This function creates a list of objects with random values for selected
        attributes from ross.DiskElement.

        Parameters
        ----------
        is_random : list
            List of the object attributes to become stochastic.
        *args : dict
            Dictionary instanciating the ross.DiskElement class.
            The attributes that are supposed to be stochastic should be
            set as lists of random variables.

        Returns
        -------
        f_list : generator
            Generator of random objects.
        """
        args_dict = args[0]
        new_args = []
        for i in range(len(args_dict[is_random[0]])):
            arg = []
            for key, value in args_dict.items():
                if key in is_random:
                    arg.append(value[i])
                else:
                    arg.append(value)
            new_args.append(arg)
        f_list = (DiskElement(*arg) for arg in new_args)

        return f_list

    def plot_random_var(self, var_list=None, histogram_kwargs=None, plot_kwargs=None):
        """Plot histogram and the PDF.

        This function creates a histogram to display the random variable
        distribution.

        Parameters
        ----------
        var_list : list, optional
            List of random variables, in string format, to plot.
            Default is plotting all the random variables.
        histogram_kwargs : dict, optional
            Additional key word arguments can be passed to change
            the plotly.go.histogram (e.g. histnorm="probability density", nbinsx=20...).
            *See Plotly API to more information.
        plot_kwargs : dict, optional
            Additional key word arguments can be passed to change the plotly go.figure
            (e.g. line=dict(width=4.0, color="royalblue"), opacity=1.0, ...).
            *See Plotly API to more information.

        Returns
        -------
        fig : Plotly graph_objects.Figure()
            A figure with the histogram plots.

        Examples
        --------
        >>> import ross.stochastic as srs
        >>> elm = srs.st_disk_example()
        >>> fig = elm.plot_random_var(["m"])
        >>> # fig.show()
        """
        label = dict(
            m="Mass",
            Id="Diametral moment of inertia",
            Ip="Polar moment of inertia",
        )
        if var_list is None:
            var_list = self.is_random
        elif not all(var in self.is_random for var in var_list):
            raise ValueError(
                "Random variable not in var_list. Select variables from {}".format(
                    self.is_random
                )
            )

        return plot_histogram(
            self.attribute_dict, label, var_list, histogram_kwargs={}, plot_kwargs={}
        )

    @classmethod
    @check_units
    def from_geometry(
        cls,
        n,
        material,
        width,
        i_d,
        o_d,
        tag=None,
        is_random=None,
    ):
        """Random disk element.

        Creates an object containing a list with random instances of
        DiskElement.from_geometry.

        Parameters
        ----------
        n: int
            Node in which the disk will be inserted.
        material: ross.Material, list of ross.Material
            Disk material.
            Input a list to make it random.
        width: float, list
            The disk width.
            Input a list to make it random.
        i_d: float, list
            Inner diameter.
            Input a list to make it random.
        o_d: float, list
            Outer diameter.
            Input a list to make it random.
        tag : str, optional
            A tag to name the element
            Default is None
        is_random : list
            List of the object attributes to become random.
            Possibilities:
                ["material", "width", "i_d", "o_d"]

        Example
        -------
        >>> import numpy as np
        >>> import ross.stochastic as srs
        >>> from ross.materials import steel
        >>> i_d=np.random.uniform(0.05, 0.06, 5)
        >>> o_d=np.random.uniform(0.35, 0.39, 5)
        >>> elms = srs.ST_DiskElement.from_geometry(n=1,
        ...                                         material=steel,
        ...                                         width=0.07,
        ...                                         i_d=i_d,
        ...                                         o_d=o_d,
        ...                                         is_random=["i_d", "o_d"],
        ...                                         )
        >>> len(list(iter(elms)))
        5
        """
        if isinstance(material, ST_Material):
            material = list(material.__iter__())
            rho = np.array([m.rho for m in material])
        else:
            rho = material.rho

        if type(width) == list:
            width = np.array(width)
        if type(i_d) == list:
            i_d = np.array(i_d)
        if type(o_d) == list:
            o_d = np.array(o_d)

        attribute_dict = dict(
            n=n,
            material=material,
            width=width,
            i_d=i_d,
            o_d=o_d,
            tag=tag,
        )
        size = len(attribute_dict[is_random[0]])

        for k, v in attribute_dict.items():
            if k not in is_random:
                v = np.full(size, v)
            else:
                v = np.array(v)

        m = 0.25 * rho * np.pi * width * (o_d**2 - i_d**2)
        # fmt: off
        Id = (
            0.015625 * rho * np.pi * width * (o_d ** 4 - i_d ** 4)
            + m * (width ** 2) / 12
        )
        # fmt: on
        Ip = 0.03125 * rho * np.pi * width * (o_d**4 - i_d**4)

        is_random = ["m", "Id", "Ip"]

        return cls(n, m, Id, Ip, tag, is_random=is_random)


def st_disk_example():
    """Return an instance of a simple random disk.

    The purpose is to make available a simple model so that doctest can be
    written using it.

    Returns
    -------
    elm : ross.stochastic.ST_DiskElement
        An instance of a random disk element object.

    Examples
    --------
    >>> import ross.stochastic as srs
    >>> elm = srs.st_disk_example()
    >>> len(list(iter(elm)))
    2
    """
    elm = ST_DiskElement(
        n=1,
        m=[30, 40],
        Id=[0.2, 0.3],
        Ip=[0.5, 0.7],
        is_random=["m", "Id", "Ip"],
    )
    return elm



from abc import ABC, abstractmethod

import numpy as np
import plotly.graph_objects as go
import scipy as sp
from warnings import warn

from ross.results import TimeResponseResults
from ross.units import Q_

__all__ = ["Fault"]


class Fault(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def run(self):
        pass

    def run_time_response(self):
        results = TimeResponseResults(
            rotor=self.rotor,
            t=self.time_vector,
            yout=self.response.T,
            xout=[],
        )
        return results

    def plot_dfft(self, probe, probe_units="rad", range_freq=None, fig=None, **kwargs):
        """Plot response in frequency domain (dFFT - discrete Fourier Transform) using Plotly.

        Parameters
        ----------
        probe : list
            List with rs.Probe objects.
        probe_units : str, option
            Units for probe orientation.
            Default is "rad".
        range_freq : list, optional
            Units for the x axis.
            Default is "Hz"
        fig : Plotly graph_objects.Figure()
            The figure object with the plot.
        kwargs : optional
            Additional key word arguments can be passed to change the plot layout only
            (e.g. width=1000, height=800, ...).
            *See Plotly Python Figure Reference for more information.

        Returns
        -------
        fig : Plotly graph_objects.Figure()
            The figure object with the plot.
        """
        if fig is None:
            fig = go.Figure()

        num_dof = self.rotor.number_dof

        for i, p in enumerate(probe):
            probe_direction = "radial"
            try:
                node = p.node
                angle = p.angle
                probe_tag = p.tag or p.get_label(i + 1)
                if p.direction == "axial":
                    if num_dof == 6:
                        probe_direction = p.direction
                    else:
                        continue
            except AttributeError:
                node = p[0]
                warn(
                    "The use of tuples in the probe argument is deprecated. Use the Probe class instead.",
                    DeprecationWarning,
                )
                try:
                    angle = Q_(p[1], probe_units).to("rad").m
                except TypeError:
                    angle = p[1]
                try:
                    probe_tag = p[2]
                except IndexError:
                    probe_tag = f"Probe {i+1} - Node {p[0]}"

            row, cols = self.response.shape
            init_step = int(2 * cols / 3)

            if probe_direction == "radial":
                dofx = num_dof * node
                dofy = num_dof * node + 1

                # fmt: off
                operator = np.array(
                    [[np.cos(angle), np.sin(angle)],
                    [-np.sin(angle), np.cos(angle)]]
                )

                _probe_resp = operator @ np.vstack((self.response[dofx, init_step:], self.response[dofy, init_step:]))
                probe_resp = _probe_resp[0,:]
                # fmt: on
            else:
                dofz = num_dof * node + 2
                probe_resp = self.response[dofz, init_step:]

            amp, freq = self._dfft(probe_resp, self.dt)

            if range_freq is not None:
                amp = amp[(freq >= range_freq[0]) & (freq <= range_freq[1])]
                freq = freq[(freq >= range_freq[0]) & (freq <= range_freq[1])]

            fig.add_trace(
                go.Scatter(
                    x=freq,
                    y=amp,
                    mode="lines",
                    name=probe_tag,
                    legendgroup=probe_tag,
                    showlegend=True,
                    hovertemplate=f"Frequency (Hz): %{{x:.2f}}<br>Amplitude (m): %{{y:.2e}}",
                )
            )

        fig.update_xaxes(title_text=f"Frequency (Hz)")
        fig.update_yaxes(title_text=f"Amplitude (m)")
        fig.update_layout(**kwargs)

        return fig

    def _dfft(self, x, dt):
        """Calculate dFFT - discrete Fourier Transform.

        Parameters
        ----------
        x : np.array
            Magnitude of the response in time domain.
            Default is "m".
        dt : int
            Time step.
            Default is "s".

        Returns
        -------
        x_amp : np.array
            Amplitude of the response in frequency domain.
            Default is "m".
        freq : np.array
            Frequency range.
            Default is "Hz".
        """
        b = np.floor(len(x) / 2)
        c = len(x)
        df = 1 / (c * dt)

        x_amp = sp.fft.fft(x)[: int(b)]
        x_amp = x_amp * 2 / c
        x_phase = np.angle(x_amp)
        x_amp = np.abs(x_amp)

        freq = np.arange(0, df * b, df)
        freq = freq[: int(b)]  # Frequency vector

        return x_amp, freq



import time

import numpy as np
from scipy import linalg as la

import ross
from ross.units import Q_, check_units

from .abs_fault import Fault
from .integrate_solver import Integrator

__all__ = [
    "Rubbing",
]


class Rubbing(Fault):
    """Contains a rubbing model for applications on finite element models of rotative machinery.
    The reference coordenates system is: z-axis throught the shaft center; x-axis and y-axis in the sensors' planes

    Parameters
    ----------
    dt : float
        Time step.
    tI : float
        Initial time.
    tF : float
        Final time.
    deltaRUB : float
        Distance between the housing and shaft surface.
    kRUB : float
        Contact stiffness.
    cRUB : float
        Contact damping.
    miRUB : float
        Friction coefficient.
    posRUB : int
        Node where the rubbing is ocurring.
    speed : float, pint.Quantity
        Operational speed of the machine. Default unit is rad/s.
    unbalance_magnitude : array
        Array with the unbalance magnitude. The unit is kg.m.
    unbalance_phase : array
        Array with the unbalance phase. The unit is rad.
    torque : bool
        Set it as True to consider the torque provided by the rubbing, by default False.
    print_progress : bool
        Set it True, to print the time iterations and the total time spent, by default False.

    Returns
    -------
    A force to be applied on the shaft.

    References
    ----------
    .. [1] Yamamoto, T., Ishida, Y., &Kirk, R.(2002). Linear and Nonlinear Rotordynamics: A Modern Treatment with Applications, pp. 215-222 ..

    Examples
    --------
    >>> from ross.probe import Probe
    >>> from ross.faults.rubbing import rubbing_example
    >>> probe1 = Probe(14, 0)
    >>> probe2 = Probe(22, 0)
    >>> response = rubbing_example()
    >>> results = response.run_time_response()
    >>> fig = response.plot_dfft(probe=[probe1, probe2], range_freq=[0, 100], yaxis_type="log")
    >>> # fig.show()
    """

    @check_units
    def __init__(
        self,
        dt,
        tI,
        tF,
        deltaRUB,
        kRUB,
        cRUB,
        miRUB,
        posRUB,
        speed,
        unbalance_magnitude,
        unbalance_phase,
        torque=False,
        print_progress=False,
    ):
        self.dt = dt
        self.tI = tI
        self.tF = tF
        self.deltaRUB = deltaRUB
        self.kRUB = kRUB
        self.cRUB = cRUB
        self.miRUB = miRUB
        self.posRUB = posRUB
        self.speed = speed
        self.speedI = speed
        self.speedF = speed
        self.DoF = np.arange((self.posRUB * 6), (self.posRUB * 6 + 6))
        self.torque = torque
        self.unbalance_magnitude = unbalance_magnitude
        self.unbalance_phase = unbalance_phase
        self.print_progress = print_progress

        if len(self.unbalance_magnitude) != len(self.unbalance_phase):
            raise Exception(
                "The unbalance magnitude vector and phase must have the same size!"
            )

    def run(self, rotor):
        """Calculates the shaft angular position and the unbalance forces at X / Y directions.

        Parameters
        ----------
        rotor : ross.Rotor Object
             6 DoF rotor model.

        """

        self.rotor = rotor
        self.n_disk = len(self.rotor.disk_elements)
        if self.n_disk != len(self.unbalance_magnitude):
            raise Exception("The number of discs and unbalances must agree!")

        self.ndof = rotor.ndof
        self.iteration = 0
        self.radius = rotor.df_shaft.iloc[self.posRUB].o_d / 2
        self.ndofd = np.zeros(len(self.rotor.disk_elements))

        for ii in range(self.n_disk):
            self.ndofd[ii] = (self.rotor.disk_elements[ii].n) * 6

        self.lambdat = 0.00001
        # Faxial = 0
        # TorqueI = 0
        # TorqueF = 0

        self.sA = (
            self.speedI * np.exp(-self.lambdat * self.tF)
            - self.speedF * np.exp(-self.lambdat * self.tI)
        ) / (np.exp(-self.lambdat * self.tF) - np.exp(-self.lambdat * self.tI))
        self.sB = (self.speedF - self.speedI) / (
            np.exp(-self.lambdat * self.tF) - np.exp(-self.lambdat * self.tI)
        )

        # sAT = (
        #     TorqueI * np.exp(-lambdat * self.tF) - TorqueF * np.exp(-lambdat * self.tI)
        # ) / (np.exp(-lambdat * self.tF) - np.exp(-lambdat * self.tI))
        # sBT = (TorqueF - TorqueI) / (
        #     np.exp(-lambdat * self.tF) - np.exp(-lambdat * self.tI)
        # )

        # self.SpeedV = sA + sB * np.exp(-lambdat * t)
        # self.TorqueV = sAT + sBT * np.exp(-lambdat * t)
        # self.AccelV = -lambdat * sB * np.exp(-lambdat * t)

        # Determining the modal matrix
        self.K = self.rotor.K(self.speed)
        self.C = self.rotor.C(self.speed)
        self.G = self.rotor.G()
        self.M = self.rotor.M(self.speed)
        self.Ksdt = self.rotor.Ksdt()

        V1, ModMat = la.eigh(self.K, self.M)

        ModMat = ModMat[:, :12]
        self.ModMat = ModMat

        # Modal transformations
        self.Mmodal = ((ModMat.T).dot(self.M)).dot(ModMat)
        self.Cmodal = ((ModMat.T).dot(self.C)).dot(ModMat)
        self.Gmodal = ((ModMat.T).dot(self.G)).dot(ModMat)
        self.Kmodal = ((ModMat.T).dot(self.K)).dot(ModMat)
        self.Ksdtmodal = ((ModMat.T).dot(self.Ksdt)).dot(ModMat)

        y0 = np.zeros(24)
        t_eval = np.arange(self.tI, self.tF + self.dt, self.dt)
        # t_eval = np.arange(self.tI, self.tF, self.dt)
        T = t_eval

        self.angular_position = (
            self.sA * T
            - (self.sB / self.lambdat) * np.exp(-self.lambdat * T)
            + (self.sB / self.lambdat)
        )

        self.Omega = self.sA + self.sB * np.exp(-self.lambdat * T)
        self.AccelV = -self.lambdat * self.sB * np.exp(-self.lambdat * T)

        self.tetaUNB = np.zeros((len(self.unbalance_phase), len(self.angular_position)))
        unbx = np.zeros(len(self.angular_position))
        unby = np.zeros(len(self.angular_position))

        FFunb = np.zeros((self.ndof, len(t_eval)))
        self.forces_rub = np.zeros((self.ndof, len(t_eval)))

        for ii in range(self.n_disk):
            self.tetaUNB[ii, :] = (
                self.angular_position + self.unbalance_phase[ii] + np.pi / 2
            )

            unbx = self.unbalance_magnitude[ii] * (self.AccelV) * (
                np.cos(self.tetaUNB[ii, :])
            ) - self.unbalance_magnitude[ii] * ((self.Omega**2)) * (
                np.sin(self.tetaUNB[ii, :])
            )

            unby = -self.unbalance_magnitude[ii] * (self.AccelV) * (
                np.sin(self.tetaUNB[ii, :])
            ) - self.unbalance_magnitude[ii] * (self.Omega**2) * (
                np.cos(self.tetaUNB[ii, :])
            )

            FFunb[int(self.ndofd[ii]), :] += unbx
            FFunb[int(self.ndofd[ii] + 1), :] += unby

        self.Funbmodal = (self.ModMat.T).dot(FFunb)

        self.inv_Mmodal = np.linalg.pinv(self.Mmodal)
        t1 = time.time()

        x = Integrator(
            self.tI,
            y0,
            self.tF,
            self.dt,
            self._equation_of_movement,
            self.print_progress,
        )
        x = x.rk4()
        t2 = time.time()
        if self.print_progress:
            print(f"Time spent: {t2-t1} s")

        self.displacement = x[:12, :]
        self.velocity = x[12:, :]
        self.time_vector = t_eval
        self.response = self.ModMat.dot(self.displacement)

    def _equation_of_movement(self, T, Y, i):
        """Calculates the displacement and velocity using state-space representation in the modal domain.

        Parameters
        ----------
        T : float
            Iteration time.
        Y : array
            Array of displacement and velocity, in the modal domain.
        i : int
            Iteration step.

        Returns
        -------
        new_Y :  array
            Array of the new displacement and velocity, in the modal domain.
        """

        positions = Y[:12]
        velocity = Y[12:]  # velocity in space state

        positionsFis = self.ModMat.dot(positions)
        velocityFis = self.ModMat.dot(velocity)

        Frub, ft = self._rub(positionsFis, velocityFis, self.Omega[i])
        self.forces_rub[:, i] = ft
        ftmodal = (self.ModMat.T).dot(ft)

        # proper equation of movement to be integrated in time
        new_V_dot = (
            ftmodal
            + self.Funbmodal[:, i]
            - ((self.Cmodal + self.Gmodal * self.Omega[i])).dot(velocity)
            - ((self.Kmodal + self.Ksdtmodal * self.AccelV[i]).dot(positions))
        ).dot(self.inv_Mmodal)

        new_X_dot = velocity

        new_Y = np.zeros(24)
        new_Y[:12] = new_X_dot
        new_Y[12:] = new_V_dot

        return new_Y

    def _rub(self, positionsFis, velocityFis, ang):
        self.F_k = np.zeros(self.ndof)
        self.F_c = np.zeros(self.ndof)
        self.F_f = np.zeros(self.ndof)

        self.y = np.concatenate((positionsFis, velocityFis))

        ii = 0 + 6 * self.posRUB  # rubbing position

        self.radial_displ_node = np.sqrt(
            self.y[ii] ** 2 + self.y[ii + 1] ** 2
        )  # radial displacement
        self.radial_displ_vel_node = np.sqrt(
            self.y[ii + self.ndof] ** 2 + self.y[ii + 1 + self.ndof] ** 2
        )  # velocity
        self.phi_angle = np.arctan2(self.y[ii + 1], self.y[ii])

        if self.radial_displ_node >= self.deltaRUB:
            self.F_k[ii] = self._stiffness_force(self.y[ii])
            self.F_k[ii + 1] = self._stiffness_force(self.y[ii + 1])
            self.F_c[ii] = self._damping_force(self.y[ii + self.ndof])
            self.F_c[ii + 1] = self._damping_force(self.y[ii + 1 + self.ndof])

            Vt = -self.y[ii + self.ndof + 1] * np.sin(self.phi_angle) + self.y[
                ii + self.ndof
            ] * np.cos(self.phi_angle)

            if Vt + ang * self.radius > 0:
                self.F_f[ii] = -self._tangential_force(self.F_k[ii], self.F_c[ii])
                self.F_f[ii + 1] = self._tangential_force(
                    self.F_k[ii + 1], self.F_c[ii + 1]
                )

                if self.torque:
                    self.F_f[ii + 5] = self._torque_force(
                        self.F_f[ii], self.F_f[ii + 1], self.y[ii]
                    )
            elif Vt + ang * self.radius < 0:
                self.F_f[ii] = self._tangential_force(self.F_k[ii], self.F_c[ii])
                self.F_f[ii + 1] = -self._tangential_force(
                    self.F_k[ii + 1], self.F_c[ii + 1]
                )

                if self.torque:
                    self.F_f[ii + 5] = self._torque_force(
                        self.F_f[ii], self.F_f[ii + 1], self.y[ii]
                    )

        return self._combine_forces(self.F_k, self.F_c, self.F_f)

    def _stiffness_force(self, y):
        """Calculates the stiffness force

        Parameters
        ----------
        y : float
            Displacement value.

        Returns
        -------
        force : numpy.float64
            Force magnitude.
        """
        force = (
            -self.kRUB
            * (self.radial_displ_node - self.deltaRUB)
            * y
            / abs(self.radial_displ_node)
        )
        return force

    def _damping_force(self, y):
        """Calculates the damping force

        Parameters
        ----------
        y : float
            Displacement value.

        Returns
        -------
        force : numpy.float64
            Force magnitude.
        """
        force = (
            -self.cRUB
            * (self.radial_displ_vel_node)
            * y
            / abs(self.radial_displ_vel_node)
        )
        return force

    def _tangential_force(self, F_k, F_c):
        """Calculates the tangential force

        Parameters
        ----------
        y : float
            Displacement value.

        Returns
        -------
        force : numpy.float64
            Force magnitude.
        """
        force = self.miRUB * (abs(F_k + F_c))
        return force

    def _torque_force(self, F_f, F_fp, y):
        """Calculates the torque force

        Parameters
        ----------
        y : float
            Displacement value.

        Returns
        -------
        force : numpy.float64
            Force magnitude.
        """
        force = self.radius * (
            np.sqrt(F_f**2 + F_fp**2) * y / abs(self.radial_displ_node)
        )
        return force

    def _combine_forces(self, F_k, F_c, F_f):
        """Mounts the final force vector.

        Parameters
        ----------
        F_k : numpy.ndarray
            Stiffness force vector.
        F_c : numpy.ndarray
            Damping force vector.
        F_f : numpy.ndarray
            Tangential force vector.

        Returns
        -------
        Frub : numpy.ndarray
            Final force vector for each degree of freedom.
        FFrub : numpy.ndarray
            Final force vector.
        """
        Frub = F_k[self.DoF] + F_c[self.DoF] + F_f[self.DoF]
        FFrub = F_k + F_c + F_f

        return Frub, FFrub

    @property
    def forces(self):
        pass


def base_rotor_example():
    """Internal routine that create an example of a rotor, to be used in
    the associated misalignment problems as a prerequisite.

    This function returns an instance of a 6 DoF rotor, with a number of
    components attached. As this is not the focus of the example here, but
    only a requisite, see the example in "rotor assembly" for additional
    information on the rotor object.

    Returns
    -------
    rotor : ross.Rotor Object
        An instance of a flexible 6 DoF rotor object.

    Examples
    --------
    >>> rotor = base_rotor_example()
    >>> rotor.Ip
    0.015118294226367068
    """
    steel2 = ross.Material(
        name="Steel", rho=7850, E=2.17e11, Poisson=0.2992610837438423
    )
    #  Rotor with 6 DoFs, with internal damping, with 10 shaft elements, 2 disks and 2 bearings.
    i_d = 0
    o_d = 0.019
    n = 33

    # fmt: off
    L = np.array(
            [0  ,  25,  64, 104, 124, 143, 175, 207, 239, 271,
            303, 335, 345, 355, 380, 408, 436, 466, 496, 526,
            556, 586, 614, 647, 657, 667, 702, 737, 772, 807,
            842, 862, 881, 914]
            )/ 1000
    # fmt: on

    L = [L[i] - L[i - 1] for i in range(1, len(L))]

    shaft_elem = [
        ross.ShaftElement6DoF(
            material=steel2,
            L=l,
            idl=i_d,
            odl=o_d,
            idr=i_d,
            odr=o_d,
            alpha=8.0501,
            beta=1.0e-5,
            rotary_inertia=True,
            shear_effects=True,
        )
        for l in L
    ]

    Id = 0.003844540885417
    Ip = 0.007513248437500

    disk0 = ross.DiskElement6DoF(n=12, m=2.6375, Id=Id, Ip=Ip)
    disk1 = ross.DiskElement6DoF(n=24, m=2.6375, Id=Id, Ip=Ip)

    kxx1 = 4.40e5
    kyy1 = 4.6114e5
    kzz = 0
    cxx1 = 27.4
    cyy1 = 2.505
    czz = 0
    kxx2 = 9.50e5
    kyy2 = 1.09e8
    cxx2 = 50.4
    cyy2 = 100.4553

    bearing0 = ross.BearingElement6DoF(
        n=4, kxx=kxx1, kyy=kyy1, cxx=cxx1, cyy=cyy1, kzz=kzz, czz=czz
    )
    bearing1 = ross.BearingElement6DoF(
        n=31, kxx=kxx2, kyy=kyy2, cxx=cxx2, cyy=cyy2, kzz=kzz, czz=czz
    )

    rotor = ross.Rotor(shaft_elem, [disk0, disk1], [bearing0, bearing1])

    return rotor


def rubbing_example():
    """Create an example of a rubbing fault.

    This function returns an instance of a rubbing fault. The purpose is to make
    available a simple model so that a doctest can be written using it.

    Returns
    -------
    rubbing : ross.Rubbing Object
        An instance of a rubbing model object.

    Examples
    --------
    >>> rubbing = rubbing_example()
    >>> rubbing.speed
    125.66370614359172
    """

    rotor = base_rotor_example()

    rubbing = rotor.run_rubbing(
        dt=0.0001,
        tI=0,
        tF=0.5,
        deltaRUB=7.95e-5,
        kRUB=1.1e6,
        cRUB=40,
        miRUB=0.3,
        posRUB=12,
        speed=Q_(1200, "RPM"),
        unbalance_magnitude=np.array([5e-4, 0]),
        unbalance_phase=np.array([-np.pi / 2, 0]),
        torque=False,
        print_progress=False,
    )

    return rubbing
